import numpy as np
import subprocess
import os
import sys
useSO = False
if sys.maxunicode == 1114111:
    useSO = True
    sys.path.append(os.path.join(os.path.dirname(__file__), '..'))
    import bs2dpython as bs2d
BL=1.

#####################################################################
# methods for energy minimization (from elenicode/patches/patches.py)
#####################################################################

def spring_energy(coords, bonds,lx=0,ly=0,shearx=None):
    if shearx is None:
        shearx=0
    
    dxs = np.array([(coords[i][0]-coords[j][0]) for i,j in bonds])
    dys = np.array([(coords[i][1]-coords[j][1]) for i,j in bonds])
    
    if lx != 0:
        dxs[np.where(dxs > lx/2.)] = dxs[np.where(dxs > lx/2.)] - lx
        dxs[np.where(dxs < -lx/2.)] = dxs[np.where(dxs < -lx/2.)] + lx
    
    if ly != 0:
        dxs[np.where(dys > ly/2.)] = dxs[np.where(dys > ly/2.)] - shearx
        dxs[np.where(dys < -ly/2.)] = dxs[np.where(dys < -ly/2.)] + shearx
        dys[np.where(dys > ly/2.)] = dys[np.where(dys > ly/2.)] - ly
        dys[np.where(dys < -ly/2.)] = dys[np.where(dys < -ly/2.)] + ly
        
    
    distances = np.sqrt(dxs**2+dys**2)
    return np.sum((distances-BL)**2)/2.
        
def emin2d_cpp_old(coords, bonds, lx=0., ly=0., shearx=None, k=1., msgs=False, endclampx=False, endclampy=False):
    np.savetxt('__default.pts',coords)
    
    bonds = np.array(bonds)
    uniquebonds = False
    if bonds.shape[1] == 2:
        np.savetxt('__default.bds',np.array(bonds),fmt='%d')
    if bonds.shape[1] == 4:
        np.savetxt('__default.bds',np.array(bonds),fmt='%d %d %lf %lf')
        uniquebonds = True
    
    if shearx is None:
        subpargs = ["/home/paulose/science/networks/code/ballspring2d/bs2d",'--k', '%2.4f' % (k), '--lx', '%2.4f' % (lx),
                            '--ly', '%2.4f' % (ly),'--noshear']
    else:
        subpargs = ["/home/paulose/science/networks/code/ballspring2d/bs2d",'--k', '%2.4f' % (k), '--lx', '%2.4f' % (lx),
                            '--ly', '%2.4f' % (ly),'--shearx', '%2.4f' % (shearx)]
    
   
    if endclampx: subpargs.append('--clampx')
    if endclampy: subpargs.append('--clampy')
    if uniquebonds: subpargs.append('--uniquebonds')

    if msgs:
        subprocess.call(subpargs)
    else: # suppress stdout and stderr
        fnull = open(os.devnull,'w')
        subprocess.call(subpargs, stdout=fnull, stderr=fnull)

    respts = np.loadtxt('Bpost_test.pts')
    resbonds = np.loadtxt('Bpost_test.bds')
        
    # cleanup
    os.remove('__default.pts')
    os.remove('__default.bds')
    os.remove('Bpost_test.pts')
    os.remove('Bpost_test.bds')
    
    return spring_energy(respts,resbonds,lx,ly,shearx)*k, respts
   

def emin2d_so(coords, bonds, lx=0., ly=0., shearx=None, k=1., msgs=False, endclampx=False, endclampy=False):
    bs = bs2d.Bs2DInstance(lx,ly)
    bs.read_points(coords*1.)
    bs.read_bonds(bonds)
    bs.optimize()
    respts = np.zeros_like(coords,dtype=float)
    bs.get_points(respts)
    
    return spring_energy(respts,bonds[:,:2],lx,ly,shearx)*k, respts
   
def emin2d_cpp(*args, **kwargs):
    if useSO:
        return emin2d_so(*args, **kwargs)
    else:
        return emin2d_cpp_old(*args,**kwargs)

