import numpy as np
from scipy import linalg
import topomesh

'''
Helper functions to manipulate eigenstates of dynamical matrix
'''

def ordered_pair(p1,p2):
    if (p1 < p2): return (p1,p2)
    else: return (p2,p1)

def overlapmatrix(estates1, estates2,normed=True):
    if normed: return np.matrix([[np.dot(e1/linalg.norm(e1),e2/linalg.norm(e2)) for e2 in estates2] for e1 in estates1])
    return np.matrix([[np.dot(e1,e2) for e2 in estates2] for e1 in estates1])

def singleprojection(mx,vectors,i):
    return np.sum([mx[i,j]*vectors[j] for j in range(len(vectors))],axis=0)
    
def projection(mx,vectors):
    return np.array([singleprojection(mx,vectors,i) for i in range(len(vectors))])

def orthogonalize(vectors):
    q,r = linalg.qr(vectors.T,mode='economic')
    return q.T
    
def overlap_estates(eigstatecoll,frame1,frame2,states=[2,3],ptlists=None,bdlists=[None,None]):
    if isinstance(eigstatecoll,basestring):
        em1, em2 = [topomesh.unpickle_each(eigstatecoll,f) for f in (frame1,frame2)]
    else: # DataSequence
        em1, em2 = [eigstatecoll.read_data(f) for f in (frame1,frame2)]
    
    em1 = em1[1]
    em2 = em2[1]
    
    if ptlists is None: return em1[states],em2[states],overlapmatrix(em1[states],em2[states])
    
    #~ if dls is not None:
        #~ ptlist1 = []
        #~ for dl in dls:
            #~ ptlist1 += point_subset(sqmeshes[frame1],bd2idx1,dl)
            #~ ptlist2 += point_subset(sqmeshes[frame2],bd2idx1,dl)
        #~ ptlist2 = estates.point_subset(sqmeshes[frame2],bd2idx2,'ve')+estates.point_subset(sqmeshes[frame2],bd2idx2,'wd')
    #~ 
    
    if ptlists is not None:
        em1substates = [vector_subset(em1[state],ptlists[0],bdlists[0]).ravel() for state in states]
        em2substates = [vector_subset(em2[state],ptlists[1],bdlists[1]).ravel() for state in states]
        
        return em1substates, em2substates, overlapmatrix(em1substates,em2substates)
        
    

def isolate_estates(es1,es2,ptlista,ptlistb):
    '''find linear combinations of eigstate1 and eigstate2 that are r
       localized to ptlist1 and ptlist2'''
    es1primea = estate_subset(es1,ptlista)
    es1primeb = estate_subset(es1,ptlistb)
    es2primea = estate_subset(es2,ptlista)
    es2primeb = estate_subset(es2,ptlistb)
    
    # project onto the states on a
    oma = overlapmatrix([es1primea,es2primea],[es1,es2])
    omb = overlapmatrix([es1primeb,es2primeb],[es1,es2])
    eiga = linalg.eigh(oma)
    eigb = linalg.eigh(omb)
    print eiga[0] # should haa lower eigenvalue close to 0
    print eigb[0] # should haa lower eigenvalue close to 0
    
    # new eigenstates
    nb = eiga[1][0,0]*es1+eiga[1][1,0]*es2 # note nb comes from the nullspace of a and vice versa
    na = eigb[1][0,0]*es1+eigb[1][1,0]*es2
    
    return na,nb


def expigamma(eigstatefile,frameseq):
    a,b,um = overlap_estates(eigstatefile,frameseq[0][0],frameseq[0][1])
    for f1, f2 in frameseq[1:]:
        a,b,om = overlap_estates(eigstatefile,f1,f2)
        um = um*om
    return um

def cyclicshift(frameseq,shift,nframes):
    return [((i+shift)%nframes,(j+shift)%nframes) for i,j in frameseq]

def gen_frame_seq(nframes,step):
    frames = range(nframes)[0::step]
    if frames[-1] == nframes-1 and step > 2: frames[-1] = 0
    else: frames.append(0)
    return [(frames[i],frames[i+1]) for i in range(len(frames)-1)]

def point_subset(basemesh,bd2idx,dlcolor='ve',maxdepth=0):
    '''return indices of points surrounding a dislocation, with a specific
       ordering of points '''
    def ordered_pair(p1,p2):
        if (p1 < p2): return (p1,p2)
        else: return (p2,p1)
        
    dlbond = basemesh.get_dl(dlcolor)
    pf = dlbond.p1
    ps = dlbond.p2

    ptlist = []
    
    # only consider the points that are on bonds connected to the '5' and '7' point of the dl
    def descend(parentidx,nbrcolor,ptlist=[],depth=0):
        if depth > maxdepth: return ptlist
        pidx = basemesh.nbrinfo[parentidx][nbrcolor]
        nextpt = bd2idx[ordered_pair(pidx, parentidx)]
        if nextpt not in ptlist: ptlist.append(nextpt)
        if nbrcolor in basemesh.nbrinfo[pidx].keys():
            return descend(pidx,nbrcolor,ptlist,depth+1)
    
    #~ for i in sorted(basemesh.nbrinfo[ps].keys()):
        #~ pn = basemesh.nbrinfo[ps][i]
        #~ midpoint = bd2idx[ordered_pair(ps, pn)]
        #~ ptlist.append(midpoint)
    
    #~ for i in sorted(basemesh.nbrinfo[pf].keys()):
        #~ pn = basemesh.nbrinfo[pf][i]
        #~ midpoint = bd2idx[ordered_pair(pf, pn)]
        #~ ptlist.append(midpoint)   
        
    #~ return ptlist


    pt2 = []
    # point on bond connecting five and seven
    midpoint = bd2idx[ordered_pair(pf, ps)]
    pt2.append(midpoint)
    for i in [k for k in sorted(basemesh.nbrinfo[pf].keys()) if k not in dlbond.color]:
        pn = basemesh.nbrinfo[ps][i]
        for nc in ['%d' % j for j in range(6)]:
            descend(pn,nc,pt2)
    #~ for i in [k for k in sorted(basemesh.nbrinfo[ps].keys()) if k not in dlbond.color]:
        #~ pn = basemesh.nbrinfo[ps][i]
        #~ for nc in ['%d' % j for j in range(6)]:
            #~ descend(pn,nc,pt2)
    
    return pt2
    

def bond_subset(basemesh,bd2idx,idxpair2bd,dlcolor='ve'):
    '''return indices of bonds surrounding a dislocation, with a specific
       ordering of bonds '''
    pfbds = dict([])
    pfbds['ns'] = (('0','1'),('1','2'),('2','3'),('3','s'),('s','0'))
    pfbds['ev'] = (('0','1'),('1','2'),('2','v'),('v','5'),('5','0'))
    pfbds['dw'] = (('0','1'),('1','w'),('w','4'),('4','5'),('5','0'))
    pfbds['sn'] = (('0','n'),('n','3'),('3','4'),('4','5'),('5','0'))
    pfbds['ve'] = (('e','2'),('2','3'),('3','4'),('4','5'),('5','e'))
    pfbds['wd'] = (('1','2'),('2','3'),('3','4'),('4','d'),('d','1'))
    
    psbds = dict([])
    psbds['ns'] = (('0','1'),('2','3'),('3','4'),('4','5'),('5','0'),('1','n'),('n','2'))
    psbds['ev'] = (('1','2'),('2','3'),('3','4'),('4','5'),('5','0'),('0','e'),('e','1'))
    psbds['dw'] = (('0','1'),('1','2'),('2','3'),('3','4'),('4','5'),('5','d'),('d','0'))
    psbds['sn'] = (('0','1'),('1','2'),('2','3'),('3','4'),('5','0'),('4','s'),('s','5'))
    psbds['ve'] = (('0','1'),('1','2'),('2','3'),('4','5'),('5','0'),('3','v'),('v','4'))
    psbds['wd'] = (('0','1'),('1','2'),('3','4'),('4','5'),('5','0'),('2','w'),('w','3'))
        
    dlbond = basemesh.get_dl(dlcolor)
    pf = dlbond.p1
    ps = dlbond.p2

    bdlist = []
    
    def collect_bonds(idxlist,p):
        for i,j in idxlist:
            p1 = basemesh.nbrinfo[p][i]
            p2 = basemesh.nbrinfo[p][j]
            midpoint1 = bd2idx[ordered_pair(p, p1)]
            midpoint2 = bd2idx[ordered_pair(p, p2)]
            midpoint3 = bd2idx[ordered_pair(p1,p2)]
            bdlist.append(idxpair2bd[ordered_pair(midpoint1,midpoint2)])
            bdlist.append(idxpair2bd[ordered_pair(midpoint2,midpoint3)])
            bdlist.append(idxpair2bd[ordered_pair(midpoint3,midpoint1)])
            
    # only consider the points that are on bonds connected to the '5' and '7' point of the dl
    collect_bonds(psbds[dlbond.color],ps)
    collect_bonds(pfbds[dlbond.color],pf)
    
    return bdlist
    
def estate_subset2(basemesh,bd2idx,eigstate,dlcolor='ve'):
    '''return eigstate with all components except those associated with
    points in point_subset to zero'''
    eigstate2 = np.zeros_like(eigstate)
    for midpoint in point_subset(basemesh,bd2idx,dlcolor): eigstate2[2*midpoint:2*midpoint+2] = eigstate[2*midpoint:2*midpoint+2]
    
    return eigstate2
    
def estate_subset(eigstate,ptlist=[],bdlist=None):
    '''return eigstate with all components except those associated with
    points in point_subset to zero'''
    eigstate2 = np.zeros_like(eigstate)
    for midpoint in ptlist: eigstate2[2*midpoint:2*midpoint+2] = eigstate[2*midpoint:2*midpoint+2]
    if bdlist is not None:
        bd0 = len(eigstate)/2
        for bdidx in bdlist: eigstate2[bd0+bdidx] = eigstate[bd0+bdidx]
    
    return eigstate2
    
def vector_subset2(basemesh,bd2idx,eigstate,dlcolor='ve'):
    '''flatten the estate_subset onto a single vector'''
    return np.array([eigstate[2*midpoint:2*midpoint+2] for midpoint in point_subset(basemesh,bd2idx,dlcolor)])
    
    
def vector_subset(eigstate,ptlist=[],bdlist=None):
    '''flatten the estate_subset onto a single vector'''
    vres = np.ravel([eigstate[2*midpoint:2*midpoint+2] for midpoint in ptlist]).tolist()
    if bdlist is not None:
        bd0 = len(eigstate)/2
        vres += [eigstate[bd0+bdidx] for bdidx in bdlist]
    return np.array(vres)

def zm_ss(eigstate,ptlist,bdlist):
    '''compare amplitudes of zero mode and self stress parts'''
    lsub = len(eigstate)/2
    zm = vector_subset(eigstate,ptlist=ptlist)
    ss = vector_subset(eigstate,bdlist=bdlist)
    return linalg.norm(zm)/(linalg.norm(zm)+linalg.norm(ss))
    

def amplitude_vs_r(eigstate,mesh,bd2idx,dl,pinpoints=None):
    eigstate = eigstate[:2*mesh.N]
    eigstatex = eigstate[::2]
    eigstatey = eigstate[1::2]
    eigstatev = np.vstack((eigstatex,eigstatey)).T
    midpoint = bd2idx[ordered_pair(dl.p1,dl.p2)]
    
    if pinpoints is not None:
        swapspots = [i for i in range(mesh.N-len(pinpoints),mesh.N) if i not in pinpoints]
        swapindices = [i for i in pinpoints if i < (mesh.N-len(pinpoints))]
        if swapspots.count(midpoint) == 1:
            midpoint = swapindices[swapspots.index(midpoint)]
        elif swapindices.count(midpoint) == 1:
            midpoint = swapspots[swapindices.index(midpoint)]
        
    origin = mesh.Points[midpoint] # centered on 5-coordinated pt of dislocation
    r = np.linalg.norm(mesh.points()-origin,axis=1)
    amplitude = np.linalg.norm(eigstatev,axis=1)
    return eigstatev,amplitude,r
    

def amplitude_vs_r_sq(eigstate,mesh,origin):
    eigstate = eigstate[:2*mesh.N]
    eigstatex = eigstate[::2]
    eigstatey = eigstate[1::2]
    eigstatev = np.vstack((eigstatex,eigstatey)).T
    r = np.linalg.norm(mesh.points()-origin,axis=1)
    amplitude = np.linalg.norm(eigstatev,axis=1)
    return eigstatev,amplitude,r
    
