"""
Input:
      all current folder's txts: with content being front position(as y, 2nd column) vs time (as x, 1st column)
Output:
      txt: front velocity vs shear rate
      fig: same above
      
"""

from scipy import stats
import numpy as np
#import matplotlib.pyplot as plt
import pylab as plt
from scipy.optimize import curve_fit
import os
from scipy.stats.distributions import  t
import matplotlib

font = {'family' : 'normal',
        'weight' : 'bold',
        'size'   : 24}
matplotlib.rc('font', **font)

outside_rangefitting=1 # 1 for range of fitting the Vf-Vs plot line, 0 for all data in the file
outside_range_min=0 # from 0
outside_range_max=4 # to the max+1: if have 6 data pts, sequence is 0 to 5, use [0:6] to get all

inside_rangefitting=0 # 1 for range of fitting of front-time to get Vf, 0 for all data in the file
inside_range_min=0
inside_range_max=12

log_log_plot=1  # if 1 loglog plot, else linear plot

dir='fit_result/' #output dir
if not os.path.exists(dir):
    os.makedirs(dir)

fout='front_shear.txt'# output txt
name='Vf-ShearRate.png'# output fig

input_tuples=[]
FileNumberMax=0
for files in os.listdir("."):
    if files.endswith(".txt"):
        print files, FileNumberMax
        plotname,ext=files.split(':')
        vins,ext2=ext.split('_')
        print vins
        vinf=float(vins)
        input_tuples.append((files,vinf,vins))
        FileNumberMax+=1

input_sorted_tuples=sorted(input_tuples,key = lambda inp:inp[1])
        

x_out=np.zeros(FileNumberMax)
slope1=np.zeros(FileNumberMax)

        
for filenumber in range(FileNumberMax):
     fileinput=input_sorted_tuples[filenumber][0]
     array1 = np.loadtxt(fileinput)
     x = array1[:,0]
     y = array1[:,1]
     if (inside_rangefitting==1):
          x1 = x[inside_range_min:inside_range_max] # set fitting range
          y1 = y[inside_range_min:inside_range_max] # set fitting range
     else:
          x1=x
          y1=y

     x_out[filenumber]=input_sorted_tuples[filenumber][1]

     slope1[filenumber], intercept, r_value, p_value, std_err = stats.linregress(x1,y1)

np.savetxt(dir+fout,np.column_stack((x_out,slope1)))

def line(t,a,b):
    return a*t+b
if (outside_rangefitting==1):
          x1 = x_out[outside_range_min:outside_range_max] # set fitting range
          y1 = slope1[outside_range_min:outside_range_max] # set fitting range
else:
          x1=x_out
          y1=slope1
logx=np.log(x1)
logy=np.log(y1)
logpars,logcovar=curve_fit(line,logx,logy)
fitted_logy = np.polyval(logpars,logx)
fitted_y=np.exp(fitted_logy)
# To get parameter confidence interval: From: http://kitchingroup.cheme.cmu.edu/blog/2013/02/12/Nonlinear-curve-fitting-with-parameter-confidence-intervals/
alpha = 0.05 # 95% confidence interval = 100*(1-alpha)
n = len(logy)    # number of data points
p = len(logpars) # number of parameters
dof = max(0, n - p) # number of degrees of freedom
tval = t.ppf(1.0-alpha/2., dof) # student-t value for the dof and confidence level
sigma=np.diag(logcovar)**0.5

plotlabel=r'$y=%0.2fx^{%0.3f\pm%0.3f}$'%(np.exp(logpars[1]),logpars[0],sigma[0]*tval)
if log_log_plot==1:
    # ##################--loglog plot
    plt.loglog(x_out,slope1,marker='o',ms=10,linestyle='None',basex=10)
    plt.loglog(x1, fitted_y, '--',label=plotlabel)
else:
    # #####################---linear scale plot
    plt.plot(x_out,slope1,marker='o',linestyle='None')
    plt.plot(x1, fitted_y, 'g--',label=plotlabel)


plt.legend(loc='best')
#plt.xlim([1e-4,1e0])


plt.xlabel('Shear Rate',fontsize=30)
plt.ylabel('Front Velocity',fontsize=30)
plt.savefig(dir+name)
plt.show()
