"""
Input: command
Output: show fig or save fig
"""

from scipy import stats
import numpy as np
import pylab as plt
from scipy.optimize import curve_fit
from scipy.stats.distributions import  t
import os
import sys
import matplotlib
from pylab import *

font = {'family' : 'normal',
        'weight' : 'normal',
        'size'   : 22}

matplotlib.rc('font', **font)

matplotlib.rc('axes', titlesize=22)
matplotlib.rc('legend', fontsize=28)
matplotlib.rc('xtick', labelsize=16)
matplotlib.rc('ytick', labelsize=16)


# font = {'family' : 'normal',
#         'weight' : 'bold',
#         'size'   : 16}
# matplotlib.rc('font', **font)

dir='fit_result/' #output dir
if not os.path.exists(dir):
    os.makedirs(dir)
outfigname='impulse_front-time.pdf'# output fig
FileNumberMax=0
input_tuples=[]
# set if 1 to enable data range and fitting range
userange=0
range_min=0
range_max=39

usefitrange=1
fit_range_min=0
fit_range_max=13

#set if 1 log plot else linear plot
log_log_plot=1

for files in os.listdir("."):
    if files.endswith(".txt"):
        # print files, FileNumberMax
        input_tuples.append((files))
        FileNumberMax+=1

input_sorted_tuples=sorted(input_tuples,key = lambda inp:inp[0])

for filenumber in range(0,FileNumberMax):
        fileinput=input_sorted_tuples[filenumber]
        # print fileinput

        input=np.loadtxt(fileinput)
        if userange==1:
            x=input[range_min:range_max,0]
            y=input[range_min:range_max,1]
        else:
            x=input[:,0]
            y=input[:,1]

        if usefitrange==1:
            xfit=input[fit_range_min:fit_range_max,0]
            yfit=input[fit_range_min:fit_range_max,1]
        else:
            xfit=input[:,0]
            yfit=input[:,1]

        # #-----------non-linear fit
        def powerlaw(t,a,b):
            return a*(t**b)
        pars,covar =  curve_fit(powerlaw,xfit,yfit)
        # #------------log-log linear fit with base e
        def line(t,a,b):
            return a*t+b
        logx=np.log(xfit)
        logy=np.log(yfit)
        #logpars=np.polyfit(logx,logy,1)
        logpars,logcovar=curve_fit(line,logx,logy)
        fitted_logy = np.polyval(logpars,logx)
        fitted_y=np.exp(fitted_logy)
        # To get parameter confidence interval: From: http://kitchingroup.cheme.cmu.edu/blog/2013/02/12/Nonlinear-curve-fitting-with-parameter-confidence-intervals/
        alpha = 0.05 # 95% confidence interval = 100*(1-alpha)
        n = len(logy)    # number of data points
        p = len(logpars) # number of parameters
        dof = max(0, n - p) # number of degrees of freedom
        tval = t.ppf(1.0-alpha/2., dof) # student-t value for the dof and confidence level
        sigma=np.diag(logcovar)**0.5

        
        # #------------------plot
        plotlabel="Simulation data"
        fitlabel=r'$Fitting\ curve:\ \frac{v_h}{v_l}=%0.2f(\frac{v_E}{v_l})^{%0.3f\pm%0.3f}$'% (np.exp(logpars[1]),logpars[0],sigma[0]*tval)# <<<<<<<<<<<---------

#        plotlabel='%s'% input_sorted_tuples[filenumber]# <<<<<<<<<<<---------
        
        if log_log_plot==1:
            # ##################### ---log-log plot
            plt.loglog(x,y,marker='o',ms=10,linestyle='None',basex=10,label=plotlabel) # plot with base 10
        #    plt.loglog(xfit,powerlaw(xfit,*pars),'r--',label=r'$nonlinear fitting curve, y=%0.2fx^{%0.3f}$'%(pars[0],pars[1]))
            plt.loglog(xfit, fitted_y, 'g--',label=fitlabel)
            
            
        else:
            # ####################---linear scale plot
            plt.plot(x,y,marker='o',ms=10, linestyle='None',label=plotlabel) # <<<<<<<<<<---------
        #    plt.plot(xfit,powerlaw(x,*pars),'r--',label=r'$nonlinear fitting curve, y=%0.2fx^%0.3f$'%(pars[0],pars[1]))
            plt.plot(xfit, fitted_y, 'g--')

# ---------------- plot setting and save
evf=arange(-3,1,1)
vf=[10**float(i) for i in evf]
vf=np.array(vf)
vs=sqrt(vf**4+2*vf**3) 
plt.loglog(vs,vf,label=r"$(\frac{v_E}{v_l})^2=(\frac{v_h}{v_l})^4+2(\frac{v_h}{v_l})^3$")


# params = {'legend.fontsize': 30,
#           'legend.linewidth': 2}
# plt.rcParams.update(params)
plt.legend(loc='best')
# plt.xlim([1e-4,1e0])
plt.xlabel(r'$\frac{v_E}{v_l}$',fontsize=40) # <<<<<<<<<<<<<<<<<<<<----------
plt.ylabel(r'$\frac{v_h}{v_l}$',fontsize=40) # <<<<<<<<<<<-------------------
fig = matplotlib.pyplot.gcf()
fig.set_size_inches(18.5, 10.5)
fig.savefig(dir+outfigname,bbox_inches='tight')
plt.show()
