//Last edit: 11.10.2013 15:18:08 By Yujie Zhou

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <time.h>
#include <fstream>
#include <iostream>
#include <iomanip>
#include <ctime>
#include <list>
using namespace std;

#define FILENAME_LENGTH 200
#define LINELENGTH 10000
#define DELIMS " ,\t"
#define MAXDIM 3
#define floats double
#define twoPi 6.2831853071795864769

#define dx0_to_b(dx0, b, TYPE) \
  switch (TYPE) {             \
    case 1:                   \
      b = dx0; break;         \
    case 2:                   \
      b = dx0; break;         \
    case 3:                   \
      b = dx0*dx0; break;     \
    case 4:                   \
      b = pow(dx0,-6); break; \
    case 5:                   \
      b = dx0; break;         \
    default:                  \
      b = dx0; break;         \
}


#define CALC_dX0(XX, iD, jD, dX)   \
      for(nu=0; nu<D; nu++) { \
        dX[nu] = XX[iD + nu] - XX[jD + nu]; \
      } 


#define CALC_dV0(VV, iD, jD, dV)   \
      for(nu=0; nu<D; nu++) { \
        dV[nu] = VV[iD + nu] - VV[jD + nu]; \
      } 

#define ADD_COMPONENTS(ADD_WHAT)          \
      for(j=0; j<Wall_AtomCount0; j++) {  \
        k  = Wall_AtomIndex0[j];          \
        kD = k*D;                         \
        for(nu=0; nu<pseudoD; nu++) {     \
          kDpnu = kD+nu;                  \
          ADD_WHAT                        \
        }                                 \
      }                                   \
                                          \
      for(wi=1; wi<=WallN; wi++) {              \
	for(j=0; j<Wall_AtomCount[wi]; j++) {   \
          k  = Wall_AtomIndex[wi][j];           \
	  if(LWallNu[k] >= 0) {                 \
	    kD    = k*D;                        \
            kDpnu = kD + LWallNu[k];            \
            ADD_WHAT                            \
          }                                     \
	}                                       \
     }				    	        \





#define ADD_BOUNDARYFORCE(FORCE_TYPE, bi, nu, Fnu) \
    if(BoundaryConnQ[bi] != 0) {  	   \
      if(BoundaryConnQ[bi] > 0)   	   \
	FORCE_TYPE[bi][nu] -= Fnu; 	   \
      else				   \
	FORCE_TYPE[bi][nu] += Fnu; 	   \
    }

#define ExecuteOutFileCmd(FileBase) \
     	if( OutFileCmdQ > 0 ) {             \
	  strcpy(OutFileCmdExe,OutFileCmd); \
	  strcat(OutFileCmdExe,addNDot(FileBase,PosOutCount).c_str()); \
	  system(OutFileCmdExe);            \
	}				    \



struct BoundaryOsciStruct {
  floats time;
  floats amplitude[MAXDIM][MAXDIM];
  floats frequency[MAXDIM][MAXDIM];
  floats timeShift[MAXDIM][MAXDIM];
};

floats dt,dt_h;
floats SQRT3DT;
floats GAMMA,GAMMAi;
floats SIGMA,SIGMAoGAMMA;
long N,D,ND;
long pseudoD = -1;
floats L[MAXDIM];
floats Lh[MAXDIM];
long NSubBoxes[MAXDIM];

floats vBoundary[MAXDIM][MAXDIM]; // current direction of the boundaries
BoundaryOsciStruct BoundaryOsci;  // current oscillation velocity of the boundaries

floats L_LE[MAXDIM][MAXDIM]; 
floats L_LEinv[MAXDIM][MAXDIM];
int LeesEdwardsTrue;
int BoundaryOsciTrue;
int OverdampedTrue = 0;
int PeriodicBoundaryTrue[MAXDIM];
floats fmod_OffsetHalf(floats x, floats y);

struct interaction {
  int i1;      // first particle of the interaction
  int i2;      // second particle of the interaction
  short type;  // interaction type
  short damp;  // damping type ( 0 for no damping )
  floats dx0;   // the actual rest length
  floats dxbb1; // bond breaks for dx < dxbb1
  floats dxbb2; // bond breaks for dx > dxbb2
  floats dx2bb1,dx2bb2; // square of the above
  floats a;     // the energy scale
  floats b;     // typically sets the rest length
  floats c;     // the damping coefficient. Only considered for interactions with damp>0.
  floats c1,c2; // damping factors: 
                // [damping force] = force_damp(c,v) * (c1*[total velocity] + c2*[normal velocity])
};

struct connectionCutStruct {
  long i1;      // first particle of the interaction
  long i2;      // second particle of the interaction
  long cutQ;   // cut connection ?
};

struct connectionClosePairs {
  long i1;      // first particle of the interaction
  long i2;      // second particle of the interaction
  floats dX2;   // distance squared
};

struct WallMoveStruct {
      floats time;
      int WallNumber;
      floats direction[MAXDIM];
};

struct BoundaryMoveStruct {
  floats time;
  floats vel[MAXDIM][MAXDIM];
};



void CALC_Xinv(floats *XX, floats *XXinv) {
  static int nu,bi;
  static floats dXinv;

  if (LeesEdwardsTrue) {
    for (bi=0; bi<D; bi++) {
      dXinv = 0.;
      for (nu=0; nu<D; nu++)
        dXinv += L_LEinv[bi][nu] * XX[nu];
      
      XXinv[bi] = dXinv - floor(dXinv);
    }
  }
  else {
    for(nu=0; nu<D; nu++) {
      if (L[nu]>0.) {
	dXinv = XX[nu]/L[nu]; 
	XXinv[nu] = dXinv - floor(dXinv);
      }
      else {
	XXinv[nu] = XX[nu];
      }
    }
  }
}

void CALC_dX(floats *dX, long *PeriodicCount, floats L_LE0[MAXDIM][MAXDIM] = L_LE, floats L_LEinv0[MAXDIM][MAXDIM] = L_LEinv) {
  static int nu,bi;
  static floats dXinv;
  
  if (LeesEdwardsTrue) {
    for (bi=0; bi<D; bi++) {
      dXinv = 0.;
      for (nu=0; nu<D; nu++) {
        dXinv += L_LEinv0[bi][nu] * dX[nu];
      }
      PeriodicCount[bi] = (long)floor(dXinv + .5);
    }
    
    for (bi=0; bi<D; bi++) {
      for (nu=0; nu<D; nu++) {
        dX[bi] -= L_LE0[bi][nu] * PeriodicCount[nu]; 
      }
    }
  }
  else {
    for(nu=0; nu<D; nu++) {
      if (L[nu]>0.) {
	dXinv = floor(dX[nu]/L[nu] + .5);
	dX[nu] = dX[nu] - L[nu]*dXinv;  // = fmod_OffsetHalf(dX[nu], L[nu]); 
	PeriodicCount[nu] = (long)dXinv;
      }
      else
	PeriodicCount[nu] = 0;
    }
  }
}

void CALC_dV(floats *dV, long *PeriodicCount) {
  static int nu,bi;
  
  if (LeesEdwardsTrue) {
    for (bi=0; bi<D; bi++) {
      for (nu=0; nu<D; nu++) {
        dV[nu] -= vBoundary[bi][nu] * PeriodicCount[nu]; 
      }
    }
  }
  else {
    for(nu=0; nu<D; nu++) 
      if (L[nu]>0.) 
        dV[nu] -= vBoundary[nu][nu] * PeriodicCount[nu]; 
  }
}

floats SQR(floats value) {
  return value*value;
}

floats CALC_SQR_VEC(floats *VEC) {
  static floats VEC2;
  static int nu;
  
  VEC2 = VEC[0]*VEC[0];
  for(nu=1;nu<D;nu++)
    VEC2 += VEC[nu]*VEC[nu];
  
  return VEC2;
}



int Inverse (double aIn[][MAXDIM], double ainv[][MAXDIM], int n) {
  int   i, j;                    // Zeile, Spalte
  int   s;                       // Elimininationsschritt
  int   pzeile;                  // Pivotzeile
  int   fehler = 0;              // Fehlerflag
  double f;                      // Multiplikationsfaktor
  const double Epsilon = 0.001;  // Genauigkeit
  double Maximum;                // Zeilenpivotisierung
  //extern FILE *fout;
  int pivot = 1;
  double a[MAXDIM][2*MAXDIM];
  

  // ergänze die Matrix a um eine Einheitsmatrix (rechts anhängen)
  for (i = 0; i < n; i++) {
    for (j = 0; j < n; j++)
      a[i][j] = aIn[i][j];

    for (j = 0; j < n; j++) {
      if (i==j) a[i][n+j] = 1.0;
      else      a[i][n+j] = 0.0;
    }
  }
#if DEBUG
  MatOut (stdout, a, n, 2*n);
#endif
  
  // die einzelnen Eliminationsschritte
  s = 0;
  do {
	  // Pivotisierung vermeidet unnötigen Abbruch bei einer Null in der Diagnonalen und
		// erhöht die Rechengenauigkeit
    Maximum = fabs(a[s][s]);
    if (pivot)
    {
      pzeile = s ; 
      for (i = s+1; i < n; i++)
        if (fabs(a[i][s]) > Maximum) {
        Maximum = fabs(a[i][s]) ;
        pzeile = i;
        }
    }
    fehler = (Maximum < Epsilon);

    if (fehler) break;           // nicht lösbar 

    if (pivot)
    {
      if (pzeile != s)  // falls erforderlich, Zeilen tauschen
      { double h;
      for (j = s ; j < 2*n; j++) {
        h = a[s][j];
        a[s][j] = a[pzeile][j];
        a[pzeile][j]= h;
      }
      }
    }

    // Eliminationszeile durch Pivot-Koeffizienten f = a[s][s] dividieren
    f = a[s][s];
    for (j = s; j < 2*n; j++)
      a[s][j] = a[s][j] / f;

    // Elimination --> Nullen in Spalte s oberhalb und unterhalb der Diagonalen
    // durch Addition der mit f multiplizierten Zeile s zur jeweiligen Zeile i
    for (i = 0; i < n; i++ ) {
      if (i != s) 
      {
        f = -a[i][s];                 // Multiplikationsfaktor
        for (j = s; j < 2*n ; j++)    // die einzelnen Spalten
          a[i][j] += f*a[s][j];       // Addition der Zeilen i, s
      }
    }
#if DEBUG
    fprintf(stdout, "Nach %1i-tem Eliminationschritt:\n", s+1);
    MatOut (stdout, a, n, 2*n);
#endif
    s++;
  } while ( s < n ) ;

  if (fehler) 
  {
    fprintf (stdout, "Inverse: Matrix ist singulär\n");
    return 0; 
  }
  // Die angehängte Einheitsmatrix Matrix hat sich jetzt in die inverse Matrix umgewandelt
  // Umkopieren auf die Zielmatrix
  for (i = 0; i < n; i++) {
    for (j = 0; j < n; j++) {
      ainv[i][j] = a[i][n+j];
    }
  }
  return 1;  
}

void PrintMatrix(floats MAT[][MAXDIM]) {
      int bi,nu; 
      for (bi=0; bi<D; bi++) {
        cout << "| ";
        for (nu=0; nu<D; nu++)
          cout << MAT[bi][nu] << "\t";
        
        cout << "|" << endl;
      }
}

void PrintOsciMatrix(BoundaryOsciStruct OSCI) {
  int bi,nu; 
  for (bi=0; bi<D; bi++) {
    cout << "|";
    for (nu=0; nu<D; nu++) {
      cout << " " << OSCI.amplitude[bi][nu] << "\tcos( 2 pi " << OSCI.frequency[bi][nu] << "\t";
      if (OSCI.timeShift[bi][nu] == 0.)
        cout << " t";
      else
        cout << "( t - " << OSCI.timeShift[bi][nu] << " )";
      cout << " )\t";
    }
    cout << "|" << endl;
  }
}

floats fmodt(floats x, floats y) {
  return x - y*floor(x/y);
}

floats fmod_OffsetHalf(floats x, floats y) {
  return x - y*floor(x/y + .5);
}

/*int fmod_LeesEdwards(floats *DeltaX) {
  static floats dX_PeriodicCount[MAXDIM];
  static int nu,bi;
  
  for (nu=0; nu<D; nu++)
    dX_PeriodicCount[nu] = floor(DeltaX[nu]/L[nu] + .5);
  
  for (nu=0; nu<D; nu++) {
    DeltaX[nu] -= L[nu] * dX_PeriodicCount[nu];
    for (bi=0; bi<D; bi++) {
      if (nu != bi)
        DeltaX[nu] -= LEshift[bi][nu] * dX_PeriodicCount[bi];
    }
  }
  
  return 0;
}*/


int comp0(const void *a,const void *b) {  //compare the first matrix elements
  return ( **(floats**)a - **(floats**)b );
}

int compWM(const void *a,const void *b) {  
  return ( (*(WallMoveStruct*)a).time > (*(WallMoveStruct*)b).time );
}

int compBM(const void *a,const void *b) {  
  return ( (*(BoundaryMoveStruct*)a).time > (*(BoundaryMoveStruct*)b).time );
}

int compBO(const void *a,const void *b) {  
  return ( (*(BoundaryOsciStruct*)a).time > (*(BoundaryOsciStruct*)b).time );
}

void CheckParticleNumber(const char *POSFILE, long *N, long *D){
  int d,k;
  floats dummy;
  ifstream POS0(POSFILE);
    
  if(!POS0.is_open()) {
    cout << "Position file \"" << POSFILE << "\" essential, but not found." << endl;
    exit(42);
  }
  
    *D=0;
    while(POS0.peek()!=13 && POS0.peek()!=10 && !POS0.eof()) {   // if next char is newline "\n" (ASCII 13) or carriage return "\r" (ASCII 10)
      POS0 >> dummy;
      (*D)++;
      //cout << dummy << "  " << *D << endl;
    }

    k=1;
    while(!POS0.eof()){
      POS0 >> dummy;
      d=1;
      while(POS0.peek()!=13 && POS0.peek()!=10 && !POS0.eof()) {
	POS0 >> dummy;
	d++;
      }
      if (d==1)
	break;
      k++;
      if (d!=(*D)) {printf("Dimension not consistent! Line %d in %s\n",k,POSFILE);exit(1);}
    }
    POS0.close();
    *N=k;
}

long CheckNumberOfLines(const char *FILE) {
  long k=0;
  char dummy[1000];
  ifstream stream(FILE);
  if (stream.is_open()) {
    while(!stream.eof()) {
      stream.getline(dummy,1000);
      //if (dummy!="")
      if ( strcmp(dummy, "") != 0)
        k++;
    }
    stream.close();
    return k;
  }
  else
    return -1;
}


string addNDot(string st, int k) {
  int kk = st.rfind(".",st.length());
  if(kk==(int)string::npos) 
    kk=st.length();
  
  char StHelp[8];
  sprintf(StHelp,"%d",k);
  return st.insert(kk,StHelp);
}

floats randomKick() {
  floats rand = drand48();
  if(rand <= .16666666667) return -SQRT3DT;
  if(rand >  .83333333333) return SQRT3DT;
  return 0.;
  
  /* *** this would be a gaussian: ***
  double z1, z2, w, y1, y2;
  do {
    z1 = 2.0 * drand48() - 1.0;
    z2 = 2.0 * drand48() - 1.0;
    w = z1 * z1 + z2 * z2;
  } while ( w >= 1.0 );
  w = sqrt( (-2.0 * log( w ) ) / w );
  y1 = z1 * w;
  //y2 = z2 * w;
  
  return(y1);
  */
}


  
floats force(short type, floats x2, floats a, floats b) {
  /** 
  * This function "force" can be used for central forces. 
  * The input parameter x2 must be x^2
  * For a potential V(x) it must return V'(x)/x
  * Hence, it is made such that the actual force vector is:   
  * [Fx Fy] = force(...) [x, y]       or   [Fx Fy Fz] = force(...) [x, y, z]
  * hence, case 1, a "constant" force(...) would correspond to a Hookian spring 
  *
  * special cases: 
  * V(x)       = b/n * x^n
  * force(...) = b   * x^(n-2) = b * x2^(n/2 - 1)
  * 
  * V(x)       = b/n * (x-x0)^n
  * force(...) = b   * (x-x0)^(n-1) / x
  **/
  static floats x2i,x6i;
  switch (type)
  {
    case 1:
      return a; // Hookian spring
      break;
    case 2: 
      return a/x2; // V(x) = a * ln(x)
      break;
    case 3: 
      return a*(b - x2);
      break;
    case 4:
      x2i = 1. / x2;
      x6i = x2i * x2i * x2i;        // Lenard Jones 
      return a*(x6i - b)*x6i*x2i;   // hence x = 1/b^(1/6) is the equilibrium distance
      break;
    case 5:
      //cout << "b = " << b << "    dx = " << sqrt(x2) << endl;
      return a * ( b/sqrt(x2) - 1. ); // Hookian Spring with rest length b
      break;

    default: 
      cout << "invalid force type: " << type << endl;
      //return 0;
      exit(1);
      break;
  }
}

floats Fcrit_to_dxbb1(short type, floats a, floats dx0, floats Fcrit) {
  //static floats x2i,x6i;
  if (Fcrit < 0.)
    return -1.;
  else {
    switch (type) {
      case 1:
        return 0.; // Hookian spring
        break;
      case 2: 
        return 0.; // V(x) = a * ln(x)
        break;
      case 3: 
        return 0.;
        break;
      case 4:
        //x2i = 1. / x2;
        //x6i = x2i * x2i * x2i;        // Lenard Jones 
        return 0.;   // hence x = 1/b^(1/6) is the equilibrium distance
        break;
      case 5:
        return dx0 - Fcrit/(2.*a); // Hookian Spring with rest length dx0
        break;
  
      default: 
        cout << "invalid force type: " << type << endl;
        //return 0;
        exit(1);
        break;
    }
  }
}

floats Fcrit_to_dxbb2(short type, floats a, floats dx0, floats Fcrit) {
  //static floats x2i,x6i;
  if (Fcrit < 0.)
    return -1.;
  else {
    switch (type) {
      case 1:
        return 0.; // Hookian spring
        break;
      case 2: 
        return 0.; // V(x) = a * ln(x)
        break;
      case 3: 
        return 0.;
        break;
      case 4:
        //x2i = 1. / x2;
        //x6i = x2i * x2i * x2i;        // Lenard Jones 
        return 0.;   // hence x = 1/b^(1/6) is the equilibrium distance
        break;
      case 5:
        return dx0 + Fcrit/(2.*a); // Hookian Spring with rest length dx0
        break;
  
      default: 
        cout << "invalid force type: " << type << endl;
        //return 0;
        exit(1);
        break;
    }
  }
}

void find_close_pairs(floats dcrit, floats *X, list<connectionClosePairs> closePairs) {
  long NBoxes[D];
  long nu;
  closePairs.clear();
  
  for(nu=0; nu<D; nu++) {
    NBoxes[nu] = (long)floor(L[nu]/dcrit);
  }
  //L_LE!

}


floats force_damp(short type, floats v2, floats c) {
  /** 
  * This function "force" can be used for central forces. 
  * The input parameter x2 must be x^2
  * For a potential V(x) it must return V'(x)/x
  * Hence, it is made such that the actual force vector is:   
  * [Fx Fy] = force(...) [x, y]       or   [Fx Fy Fz] = force(...) [x, y, z]
  * hence, case 1, a "constant" force(...) would correspond to a Hookian spring 
  *
  * special cases: 
  * V(x)       = b/n * x^n
  * force(...) = b   * x^(n-2) = b * x2^(n/2 - 1)
  * 
  * V(x)       = b/n * (x-x0)^n
  * force(...) = b   * (x-x0)^(n-1) / x
  **/
  switch (type)
  {
    case 1:
      return -c; // corresponds to F_damp(v) = c*v;
      break;

    default: 
      cout << "invalid damping force type: " << type << endl;
      exit(1);
      break;
  }
}

void calc_F(floats *X, floats *V, list<interaction> connections, floats *F, floats L_LE0[MAXDIM][MAXDIM] = L_LE, floats L_LEinv0[MAXDIM][MAXDIM] = L_LEinv) {
  long k,j,nu;
  long conn_i1D, conn_i2D;
  //long kD;
  long  dX_PeriodicCount[D];
  short conn_damp;
  floats dX[D];
  floats dV[D];
  floats dX2,dV2;
  floats Fh,Fnu,c1Fh,c2Fh,dXdV;
  
  interaction conn;
  list<interaction>::iterator it;
  
  // initialize all forces
  for(k=0; k<ND; k++)
    F[k]=0.;
  
  
  j=0;
  for(it = connections.begin(); it != connections.end(); ++it) {
    conn = *it;
    conn_i1D = conn.i1*D;
    conn_i2D = conn.i2*D;
  
    CALC_dX0(X, conn_i1D, conn_i2D, dX);
    CALC_dX(dX, dX_PeriodicCount, L_LE0, L_LEinv0);
    dX2 = CALC_SQR_VEC(dX);
  
    /*if(dX2>2500.) {
      cout << j << " ";
      cout << conn.i1 << ": [" << X[conn_i1D] << ", " << X[conn_i1D+1] << "] ";
      cout << conn.i2 << ": [" << X[conn_i2D] << ", " << X[conn_i2D+1] << "] ";
      cout << " [" << dX2 << "]" << endl;
    }
      j++;*/
  
    // calculate contribution of central force
    Fh = force(conn.type, dX2, conn.a, conn.b);
    for(nu=0; nu<D; nu++) {
      Fnu = dX[nu]*Fh;
      F[conn_i1D + nu] += Fnu;
      F[conn_i2D + nu] -= Fnu;
    }
  
    // calculate contribution of damping force
    conn_damp = conn.damp;
    if(conn_damp > 0) {
      CALC_dV0(V, conn_i1D, conn_i2D, dV);
      CALC_dV(dV, dX_PeriodicCount);
  
      if(conn_damp == 1) {
        Fh = - conn.c;
      }
      else {
        dV2 = CALC_SQR_VEC(dV);
        Fh = force_damp(conn_damp, dV2, conn.c);
      }
  
      if (conn.c1 != 0.) {
        c1Fh = conn.c1*Fh;
        for(nu=0; nu<D; nu++) {
          Fnu = dV[nu]*c1Fh;
          F[conn_i1D + nu] += Fnu;
          F[conn_i2D + nu] -= Fnu;
        }
      }
  
      if (conn.c2 != 0.) {
        dXdV = dX[0]*dV[0];
        for(nu=1; nu<D; nu++)
          dXdV += dX[nu]*dV[nu];
  
        c2Fh = conn.c2 * dXdV / dX2;
        for(nu=0; nu<D; nu++) {
          Fnu = dX[nu]*c2Fh;
          F[conn_i1D + nu] += Fnu;
          F[conn_i2D + nu] -= Fnu;
        }
      }
  
    }
  }
}


int main(int argc, char** argv){
  //if(argc!=2 && argc!=4){
  if( argc % 2 != 0 ) {
    printf("usage:\n");
    printf("./%s <Settings-File>\n\n",argv[0]);
    
    printf("<Settings-File> is a file that consequtively contains:\n");
    printf("1. File-Name for Positions of Atoms\n");
    printf("   must be lines containing continuouus particle numbers and particle coordinates and velocities\n");
    printf("2. \n");
    printf(".\n");
    printf(".\n");
    //WALLaFILE="lll.l";
    //cout addNDot(WALLaFILE,2) << "   " << WALLaFILE << endl;
    exit(1);
  }
  
  
  // ******************* prepare variables for Settings-File ******************* 
  char PosFILE[FILENAME_LENGTH];     strcpy(PosFILE, "POS.txt");
  char VelFILE[FILENAME_LENGTH];     strcpy(VelFILE, "VEL.txt");
  char ExtWallFILE[FILENAME_LENGTH]; strcpy(ExtWallFILE,  "EXT.txt");
  char ConnFILE[FILENAME_LENGTH];    strcpy(ConnFILE,     "CONN.txt");
  char InputFOLDER[FILENAME_LENGTH]; strcpy(InputFOLDER, ""); short UseInputFOLDER = 0;
  char InputFOLDERsav[FILENAME_LENGTH]; 
  
  char PosOutFILE[FILENAME_LENGTH];     strcpy(PosOutFILE,   "run/OUTx.txt");
  char VelOutFILE[FILENAME_LENGTH];     strcpy(VelOutFILE,   "run/OUTv.txt");
  char VelFluctOutFILE[FILENAME_LENGTH];strcpy(VelFluctOutFILE,"none");
  char ConnOutFILE[FILENAME_LENGTH];    strcpy(ConnOutFILE,  "run/OUTc.txt");
  char SigmaOutFILE[FILENAME_LENGTH];   strcpy(SigmaOutFILE, "none");
  char OutFileCmd[FILENAME_LENGTH];     strcpy(OutFileCmd,   "#"); short OutFileCmdQ = 0;
  char OutFileCmdExe[2*FILENAME_LENGTH];strcpy(OutFileCmdExe,"#");
  
  
  char BoundaryMoveFILE[FILENAME_LENGTH];    strcpy(BoundaryMoveFILE   , "BOUNDARYm.txt");
  char BoundaryOsciFILE[FILENAME_LENGTH];    strcpy(BoundaryOsciFILE   , "BOUNDARYo.txt");
  char WallMoveFILE[FILENAME_LENGTH];        strcpy(WallMoveFILE       , "WALLm.txt");
  char BoundaryPosOutFILE[FILENAME_LENGTH];   strcpy(BoundaryPosOutFILE  , "run/BOUNDARYpos.txt");
  char BoundaryForceOutFILE[FILENAME_LENGTH]; strcpy(BoundaryForceOutFILE, "run/BOUNDARYf.txt");
  char WallForceOutFILE[FILENAME_LENGTH];    strcpy(WallForceOutFILE   , "run/WALLf.txt");
  char WallPosOutFILE[FILENAME_LENGTH];      strcpy(WallPosOutFILE     , "run/WALLpos.txt");
  
  char IntegrationMethod[LINELENGTH];               strcpy(IntegrationMethod  , "default");
  char cuttingMethod[LINELENGTH];                   strcpy(cuttingMethod      , "moreThan");
  long cuttingMoreThanCoordNum = 3;
  floats AvCoordNum = -1.;
  int Lnu = 0;
  
  short useRandomSeedQ = 0;
  long RandomSeed = 0;
  char RandomSeedString[LINELENGTH];
  
  int    weakSpringsQ = 0;
  floats weakSpringsa = 0.;
  floats weakSpringsc = 0.;
  
  long continueAt = -1;
  char lineIN[LINELENGTH];
  char *split;
  

  floats TMAX;
  floats dt_pos, dt_force;
  floats dX2Add;
  floats dX2Remove;
  floats AddRate; long AddNumber;
  floats RemoveRate,RemoveProb;
  floats Fmax, Fmax2;
  floats defaultDamping = 0.1;
  long OUTPRECISION = 6;
  short preventOverwriteQ = 0;
  
  
  long i, j, k, nu;
  long conn_i1, conn_i2;
  long conn_i1D,conn_i2D, kD, kDpnu;
  
  for(nu=0; nu<MAXDIM; nu++)
    NSubBoxes[nu] = 1;
  
  
  
  
  // ******************* read Settings-File ******************* 
  ifstream SET(argv[1]);
  if (SET.is_open()) {
    cout << ":::::::::::::::::::::::::: reading input file " << argv[1] << " ::::::::::::::::::::::::::" << endl;
    cout << "   contents: " << endl;
    k = 0;
    while(!SET.eof()) {
      
      SET.getline(lineIN, LINELENGTH);
      //cout << lineIN << endl;
      k++;
      cout << ":" << k << ":\t";
      if ((split = strtok (lineIN, DELIMS)) != NULL ) {
        
        //cout << split << endl;
	
        if (strcmp(split, "InFilePos") == 0) {
          strcpy(PosFILE, strtok (NULL, DELIMS));
          cout << "File with initial particle positions: \t" << PosFILE;
        }
        
        else if (strcmp(split, "InFileVel") == 0) {
          strcpy(VelFILE, strtok (NULL, DELIMS)); 
          cout << "File with initial particle velocities: \t" << VelFILE;
        }
        
        else if (strcmp(split, "InFileConn") == 0) {
          strcpy(ConnFILE, strtok (NULL, DELIMS)); 
          cout << "File with initial particle bonds: \t" << ConnFILE;
        }
        
        else if (strcmp(split, "InFileExt") == 0) {
          strcpy(ExtWallFILE, strtok (NULL, DELIMS)); 
          cout << "File with extra particle properties: \t" << ExtWallFILE;
        }
	
	else if (strcmp(split, "InFolder") == 0) {
          strcpy(InputFOLDER, strtok (NULL, DELIMS)); 
          cout << "Using input folder for pos-, vel-, conn-, and ext-file: " << InputFOLDER;
	  UseInputFOLDER = 1;
        }
	
        
        else if (strcmp(split, "OutFilePos") == 0) {
          strcpy(PosOutFILE, strtok (NULL, DELIMS)); 
          cout << "Output file for particle positions: \t" << PosOutFILE;
        }
        
        else if (strcmp(split, "OutFileVel") == 0) {
          strcpy(VelOutFILE, strtok (NULL, DELIMS)); 
          cout << "Output file for particle velocities: \t" << VelOutFILE;
        }

        else if (strcmp(split, "OutFileVFluct") == 0) {
          strcpy(VelFluctOutFILE, strtok (NULL, DELIMS)); 
          cout << "Output file for particle velocities relative to mean field: \t" << VelFluctOutFILE;
        }

        else if (strcmp(split, "OutFileConn") == 0) {
          strcpy(ConnOutFILE, strtok (NULL, DELIMS)); 
          cout << "Output file for particle bonds: \t" << ConnOutFILE;
        }
        
        else if (strcmp(split, "OutFileSigma") == 0) {
          strcpy(SigmaOutFILE, strtok (NULL, DELIMS));
	  if (strcmp(SigmaOutFILE, "none") == 0) cout << "OutFileSigma = none: no output of local stress tensor";
          else 				  	 cout << "Output file for local stress tensor: \t" << SigmaOutFILE;
        }
        
        else if (strcmp(split, "InFileBoundaryMove") == 0) {
          strcpy(BoundaryMoveFILE, strtok (NULL, DELIMS)); 
          cout << "Input  file for boundary moves: \t" << BoundaryMoveFILE;
        }
        
        else if (strcmp(split, "InFileBoundaryOsci") == 0) {
          strcpy(BoundaryOsciFILE, strtok (NULL, DELIMS)); 
          cout << "Input  file for boundary oscillations: \t" << BoundaryOsciFILE;
        }
        
        else if (strcmp(split, "OutFileBoundaryPos") == 0) {
          strcpy(BoundaryPosOutFILE, strtok(NULL, DELIMS));
          cout << "Output file for pos of boundaries: \t" << BoundaryPosOutFILE;
        }
        
        else if (strcmp(split, "OutFileBoundaryForce") == 0) {
          strcpy(BoundaryForceOutFILE, strtok(NULL, DELIMS));
          cout << "Output file for forces on boundaries: \t" << BoundaryForceOutFILE;
        }
        
        else if (strcmp(split, "InFileWallMove") == 0) {
          strcpy(WallMoveFILE, strtok(NULL, DELIMS));
          cout << "Input  file for wall moves: \t\t" << WallMoveFILE;
        }
        
        else if (strcmp(split, "OutFileWallForce") == 0) {
          strcpy(WallForceOutFILE, strtok(NULL, DELIMS));
          cout << "Output file for forces on walls: \t" << WallForceOutFILE;
        }
        
        else if (strcmp(split, "OutFileWallPos") == 0) {
          strcpy(WallPosOutFILE, strtok(NULL, DELIMS));
          cout << "Output file for positions of walls: \t" << WallPosOutFILE;
        }

        else if (strcmp(split, "dt") == 0) {
          dt = atof(strtok(NULL, DELIMS));
          dt_h = dt*0.5;
          cout << "dt =\t" << dt;
        }
        
        else if (strcmp(split, "tmax") == 0) {
          TMAX = atof(strtok(NULL, DELIMS));
          cout << "tmax =\t" << TMAX;
        }
        
        else if (strcmp(split, "gamma") == 0) {
          GAMMA  = atof(strtok(NULL, DELIMS));
          cout << "gamma =\t" << GAMMA;
        }
        
        else if (strcmp(split, "sigma") == 0) {
          SIGMA = atof(strtok(NULL, DELIMS));
          cout << "sigma =\t" << SIGMA;
        }
        
        else if (strcmp(split, "dXAdd") == 0) {
          dX2Add = atof(strtok(NULL, DELIMS));
          cout << "maximum particle distance for adding bonds (dXAdd) =\t" << dX2Add;
          dX2Add *= dX2Add;
          cout << "\t\tdXAdd^2 = " << dX2Add;
        }
        
        else if (strcmp(split, "dXRemove") == 0) {
          dX2Remove = atof(strtok(NULL, DELIMS));
          cout << "particle distance beyond which bonds break (dXRemove) =\t" << dX2Remove;
          dX2Remove *= dX2Remove;
          cout << "\t\tdXRemove^2 = " << dX2Remove;
        }
        
        else if (strcmp(split, "FRemove") == 0) {
          Fmax = atof(strtok(NULL, DELIMS));
          if (Fmax >= 0.) {
            cout << "Force beyond which bonds break (FRemove) =\t" << Fmax;
            Fmax2 = Fmax*Fmax;
          }
          else {
            cout << "Bonds do not break due to maximum force.";
            Fmax2 = -1.;
          }
        }
        
        else if (strcmp(split, "AddRate") == 0) {
          AddRate   = atof(strtok(NULL, DELIMS));
          AddNumber = (long)(AddRate*dt);
          cout << "rate, at which particles closer than dXAdd are bonded =\t" << AddRate;
          cout << "   --> AddNumber per time step = " << AddNumber;
        }
        
        else if (strcmp(split, "RemoveRate") == 0) {
          RemoveRate = atof(strtok(NULL, DELIMS));
          RemoveProb = RemoveRate*dt;
          cout << "rate, at which particles bonds break spontaneously =\t" << RemoveRate;
          cout << "   --> RemoveProb per time step = " << RemoveProb;
        }
        
        else if (strcmp(split, "DampingDef") == 0) {
          defaultDamping = atof(strtok(NULL, DELIMS));
          cout << "default damping coefficient (if entry 7 in \"" << ConnFILE << "\" is empty) c = " << defaultDamping;
        }
	
        else if (strcmp(split, "AvCoordNum") == 0) {
          AvCoordNum = atof(strtok(NULL, DELIMS));
	  if(AvCoordNum >= 0.) 
	    cout << "bonds are randomly cut to reach average coordination number: z = " << AvCoordNum;
	  else
	    cout << "AvCoordNum < 0: no bonds being cut";
        }
        
        else if (strcmp(split, "cutMethod") == 0) {
          strcpy(cuttingMethod, strtok(NULL, DELIMS));
          cout << "Cutting Method: \t";
          if(strcmp(cuttingMethod, "moreThan") == 0) {
	    cuttingMoreThanCoordNum = atol( strtok(NULL, DELIMS) );
            cout << "cut bonds with coordination number more than " << cuttingMoreThanCoordNum;
	  }

          else if(strcmp(cuttingMethod, "highestFirst") == 0) {
            cout << "cut bonds with highest coordination first";
	  }

          else {
            strcpy(cuttingMethod, "moreThan");
	    cuttingMoreThanCoordNum = 3;
            cout << split << " not recognized. Using default method: \"moreThan 3\"";
          }
        }
	
        else if (strcmp(split, "weakSprings") == 0) {
	  if( (split = strtok(NULL, DELIMS)) != NULL) weakSpringsa = atof(split);
	  if( (split = strtok(NULL, DELIMS)) != NULL) weakSpringsc = atof(split);
	  if( weakSpringsa != 0. || weakSpringsc != 0. ) {
	    weakSpringsQ = 1;
	    cout << "Instead of cutting bonds: use \"weak springs\" with spring constant a = " << weakSpringsa << ", and damping coefficient c = " << weakSpringsc << ".";
	  }
	  else {
	    cout << "Cutting bonds. Not using \"weak springs\".";
	  }
        }
        
	
        else if (strcmp(split, "dt_pos") == 0) {
          dt_pos = atof(strtok(NULL, DELIMS));
          cout << "Time step for output of particle data: dt_pos = " << dt_pos;
        }
        
        else if (strcmp(split, "dt_force") == 0) {
          dt_force = atof(strtok(NULL, DELIMS));
          cout << "Time step for output of force data:  dt_force = " << dt_force;
        }
        
        else if (strcmp(split, "BoxSize") == 0) {
          nu = 0;
          cout << "Box Size:   ";
          while( (split = strtok(NULL, DELIMS)) != NULL) {
            L[nu]  = atof( split );
            Lh[nu] = L[nu] * 0.5;
            cout << "L[" << nu << "] = " << L[nu] << "; ";
            nu++;
          }
        }
	
	else if (strcmp(split, "SubBoxes") == 0) {
          nu = 0;
          cout << "Number of subboxes:   ";
          while( (split = strtok(NULL, DELIMS)) != NULL) {
            NSubBoxes[nu]  = atol( split );
            cout << "N[" << nu << "] = " << NSubBoxes[nu] << "; ";
            nu++;
          }
        }
	
	else if (strcmp(split, "pseudoD") == 0) {
          pseudoD = atol(strtok(NULL, DELIMS));
          cout << "pseudo dimension: " << pseudoD;
        }
	
        else if (strcmp(split, "method") == 0) {
          strcpy(IntegrationMethod, strtok(NULL, DELIMS));
          cout << "Integration Method: \t";
          if(strcmp(IntegrationMethod, "LeapfrogNoDamp") == 0) 
            cout << "Leapfrog WITHOUT efficient damping";

          else if(strcmp(IntegrationMethod, "Leapfrog") == 0) 
            cout << "Leapfrog with efficient damping";

          else if(strcmp(IntegrationMethod, "VelocityVerlet") == 0) 
            cout << "Velocity Verlet with efficient damping";

          else if(strcmp(IntegrationMethod, "Euler") == 0) 
            cout << "Euler integration";

          else if(strcmp(IntegrationMethod, "Overdamped") == 0) {
            cout << "Overdamped dynamics, using Heun method. Particle masses are ignored.";
	    OverdampedTrue = 1;
	  }

          else {
            strcpy(IntegrationMethod, "Heun");
            cout << "default algorithm: Heun method";
          }
        }
	
	else if (strcmp(split, "outPrecisn") == 0) {
          OUTPRECISION = atol( strtok(NULL, DELIMS) );
          cout << "Precision (number of digits) for pos, vel and conn output files: " << OUTPRECISION;
        }
	
	else if (strcmp(split, "randomSeed") == 0) {
          useRandomSeedQ = 1;
	  strcpy(RandomSeedString, strtok(NULL, DELIMS));
          if(strcmp(RandomSeedString, "time") == 0) {
	    RandomSeed = time(NULL);
	    cout << "Using random seed based on current time: " << RandomSeed;
	  }
	  else {
	    RandomSeed = atol( RandomSeedString );
	    cout << "Using random seed: " << RandomSeed;
	  }
        }
	
	else if (strcmp(split, "LWallNu") == 0) {
          Lnu = atol( strtok(NULL, DELIMS) );
          cout << "dimension which is not fixed for walls: " << Lnu;
        }
        
        else if (strcmp(split, "OutFileCmd") == 0) {
          strcpy(OutFileCmd, strtok(NULL, ""));
	  strcat(OutFileCmd, " ");
	  OutFileCmdQ = 1;
          cout << "Post-processing large output files with the command: \t" << OutFileCmd << "<filename>";
        }

	
	else if (strcmp(split, "continueAt") == 0) {
          continueAt = atol( strtok(NULL, DELIMS) );
	  if(continueAt >= 0) {
	    cout << "Trying to continue the current run at output file # " << continueAt;
	    preventOverwriteQ = 1;
	  }
	  else
	    cout << "starting a new run at time 0";
        }


        else {
          cout << "keyword \"" << split << "\" ignored.";
        }
        
        
      }
      cout << endl;
    
    }
    SET.close();
    cout << "::::::::::::::::::::::::::        reading done       ::::::::::::::::::::::::::" << endl << endl;
  }
  if (useRandomSeedQ == 1)
    srand48(RandomSeed);
  
  if( UseInputFOLDER == 1 ) {
    if( strcmp(PosFILE,     "none") != 0 ) {strcpy(InputFOLDERsav, InputFOLDER); strcpy(PosFILE,    strcat(InputFOLDERsav, PosFILE));}
    if( strcmp(VelFILE,     "none") != 0 ) {strcpy(InputFOLDERsav, InputFOLDER); strcpy(VelFILE,    strcat(InputFOLDERsav, VelFILE));}
    if( strcmp(ConnFILE,    "none") != 0 ) {strcpy(InputFOLDERsav, InputFOLDER); strcpy(ConnFILE,   strcat(InputFOLDERsav, ConnFILE));}
    if( strcmp(ExtWallFILE, "none") != 0 ) {strcpy(InputFOLDERsav, InputFOLDER); strcpy(ExtWallFILE,strcat(InputFOLDERsav, ExtWallFILE));}
    cout << "full input pos-file:  " << PosFILE << endl;
    cout << "full input conn-file: " << ConnFILE << endl;
  }

  
  if (argc >= 4){
    
    for (long argcounter = 2; argcounter < argc; argcounter += 2) {
      cout << "Overwriting the following settings from Settings-file \"" << argv[1] << "\":" << endl;
      if(strcmp(argv[argcounter], "continueAt") == 0) {
	continueAt = atol( argv[argcounter+1] );
	cout << "\"continueAt\": ";
	if(continueAt >= 0) {
	  cout << "Trying to continue the current run at output file # " << continueAt << endl;
	  preventOverwriteQ = 1;
	}
	else
	  cout << "starting a new run at time 0." << endl;
	//cout << "\"continueAt\"-enrty in " << argv[1] << "is being overwritten." << endl;
      }
      
      else if(strcmp(argv[argcounter], "tmax") == 0) {
	  TMAX = atof( argv[argcounter+1] );
          cout << "\"tmax\": tmax =\t" << TMAX << endl;
      }
    }
  }
  
  

  if(continueAt >= 0) {
    strcpy(PosFILE,  addNDot(PosOutFILE,continueAt).c_str());  cout << "Changing InFilePos to:  " << PosFILE << endl;
    strcpy(VelFILE,  addNDot(VelOutFILE,continueAt).c_str());  cout << "Changing InFileVel to:  " << VelFILE << endl;
    strcpy(ConnFILE, addNDot(ConnOutFILE,continueAt).c_str()); cout << "Changing InFileConn to: " << ConnFILE << endl;
    cout << endl;
  }

  CheckParticleNumber(PosFILE, &N, &D);
  ND=N*D;
  cout << endl;
  cout << "Summary:" << endl;
  cout << "========" << endl;
  cout << "Number of particles: " << N << endl;
  cout << "Spatial dimension:   " << D << endl;
  
  
  cout << "System Size:         L    = (" << L[0];
  for(nu=1; nu<D; nu++) cout << ", " << L[nu];
  cout << ")" << endl;
  
  if(0 <= pseudoD && pseudoD <= D)
    cout << "pseudo-dimension = " << pseudoD << ". POS- and VEL-components [nu > " << pseudoD << "] are fixed." << endl;
  else
    pseudoD = D;
  
  /*cout << "Sub Boxes:           N_sb = (" << NSubBoxes[0];
  for(nu=1; nu<D; nu++) cout << ", " << NSubBoxes[nu];
  cout << ")  (currently ignored)" << endl;*/
  cout << endl;
  
  if(AvCoordNum >= 0.) {
    cout << "bonds are randomly cut to reach average coordination number: z = " << AvCoordNum << endl;
    cout << "Cutting Method: \t";
    if(strcmp(cuttingMethod, "moreThan") == 0) 
      cout << "cut bonds with participants' coordination number more than " << cuttingMoreThanCoordNum << "." << endl;
    else if(strcmp(cuttingMethod, "highestFirst") == 0) 
      cout << "cut bonds with highest coordination first"  << endl << endl;
  }
  else
    cout << "no bonds being cut" << endl << endl;

  cout << "Trying to add " << AddNumber << " connections per time step" << endl;
  cout << "Removing connections with probability " << RemoveProb << " each time step" << endl << endl;
  
  
  // ******************* initialize variables ******************* 
  floats t;
  floats nextPosOut_t, nextForceOut_t;
  long count;
  long PosOutCount, ForceOutCount;

  SQRT3DT = sqrt(3.*dt);
  floats ONEminusGAMMAdt   = 1. - GAMMA*dt;
  floats ONEminusGAMMAdt_h = 1. - GAMMA*dt_h;
  //floats dtdtHalf = dt*dt*0.5;
  //floats SIGMAodt = SIGMA / dt;
  
  

  floats *X 	     = new floats[ND];	 // positions
  floats *V 	     = new floats[ND];	 // velocities
  floats *Vrand      = new floats[ND];	 // random velocity kicks
  floats *previous_X = new floats[ND];	 // positions from the previous time step
  floats *mi 	     = new floats[N];	 // inverse masses
  floats *LWallNu    = new floats[N];	 // marker and direction for loose walls
  floats dX[D];
  floats dV[D];
  
  
  long dX_PeriodicCount[D];  // count how many periodic boundaries are crossed
  int  BoundaryConnQ[D];     // stores, whether the connection crosses the system boundary
  floats VplusAccDtH;
  floats dX2,dX0;
  floats dV2;
  floats dxbb;
  floats Fh,Fnu,c1Fh,c2Fh,dXdV;
  short conn_damp;
  
  floats *R1_X = new floats[ND];
  floats *R1_V = new floats[ND];
  floats *R1_F = new floats[ND];
  floats *R2_X = new floats[ND];
  floats *R2_V = new floats[ND];
  floats *R2_F = new floats[ND];
  
  

  floats X1_inv[D];
  floats X2_inv[D];
  floats dX_inv;
  
  int wi,wiD;
  int bi;//,biD;
  int Wallnk;
  int *Walln = new int[N];
  int  WallN = 0; //number of walls
  long looseWallCount = 0;
  floats vWall[D]; //direction of a wall
  floats BoundaryForceElas[D][D];   //force on boundary ("stress tensor"), elastic contribution
  floats BoundaryForceDiss[D][D];   //force on boundary ("stress tensor"), dissipativ contribution
  floats SystemForceElas[D][D];     //summed forces ("stress tensor") in the system, elastic contribution
  floats SystemForceDiss[D][D];     //summed forces ("stress tensor") in the system, dissipativ contribution
  //floats SystemForceKine[D][D];     //summed forces ("stress tensor") in the system, kinetic contribution
  
  long Nsigma;
  if(strcmp(SigmaOutFILE, "none") == 0) Nsigma = 1;
  else 				 	Nsigma = N;

  floats ***sigmaLocalElas = new floats**[Nsigma];
  floats ***sigmaLocalDiss = new floats**[Nsigma];
  for(i=0; i<Nsigma; i++) {
    sigmaLocalElas[i] = new floats*[D];
    sigmaLocalDiss[i] = new floats*[D];
    for(j=0; j<D; j++) {
      sigmaLocalElas[i][j] = new floats[D];
      sigmaLocalDiss[i][j] = new floats[D];
    }
  }
  
  
  long NvNeighbors;
  if(strcmp(VelFluctOutFILE, "none") == 0) NvNeighbors = 1;
  else 				 	   NvNeighbors = N;
  long   *NNeighors     = new   long[NvNeighbors];
  floats *vNeighborsSum = new floats[NvNeighbors*D];
  floats  vField[D];


  long NConns     = 0;
  long NConns0    = 0;
  long NConnsGoal = -1;
  
  
  for (bi=0; bi<D; bi++) {
    for (nu=0; nu<D; nu++) {
      vBoundary[bi][nu] = 0.;
      if(bi==nu) L_LE[bi][nu] = L[nu];
      else       L_LE[bi][nu] = 0.;
      
      BoundaryOsci.amplitude[bi][nu] = 0.;
      BoundaryOsci.frequency[bi][nu] = 0.;
      BoundaryOsci.timeShift[bi][nu] = 0.;
    }
  }
  
  Inverse( L_LE, L_LEinv, D );
  
  
  interaction conn;
  list<interaction> connections;
  list<interaction>::iterator it;
  
  
  // ******************* read initial Position and Velocities ******************* 
  ifstream POS(PosFILE);
  for(k=0;k<ND;k++) 
	POS >> X[k];
  POS.close();
  
  
  
  if( OverdampedTrue ) {
    cout << "velocity file \"" << VelFILE << "\" ignored due to use of overdamped dynamics." << endl;
    for(k=0;k<ND;k++)
      V[k] = 0.;
    
    if(GAMMA == 0.) {
      cout << "Warning: gamma = 0 invalid for overdamped dynamics. Setting gamma = 1." << endl;
      GAMMA = 1.;
    }
    GAMMAi      = 1./GAMMA;
    SIGMAoGAMMA = SIGMA/GAMMA;
    
  }
  else {
    ifstream VEL(VelFILE);
    if (VEL.is_open()) {
      for(k=0;k<ND;k++)
	    VEL >> V[k];
      VEL.close();
    }
    else {
      cout << "velocity file \"" << VelFILE << "\" not found. Setting initital velocities to 0." << endl;
      for(k=0;k<ND;k++)
	V[k] = 0.;
    }
  }
  
  
  if(SIGMA==0.) {
    cout << "Sigma = 0. No temperature kicks." << endl;
    for(k=0;k<ND;k++)
      Vrand[k] = 0.;
  }
  
  
  
  
  

  // ******************* read Masses and Wall-Atoms ******************* 
  for(k=0; k<N; k++)
    LWallNu[k] = -1;

  ifstream EXTRAS(ExtWallFILE);
  if(EXTRAS.is_open()) {
    for(k=0; k<N; k++) {
      EXTRAS >> Wallnk;
      EXTRAS >> mi[k];   	// inverse mass
      
      if (Wallnk>0)
	mi[k] = 0.;

      if(Wallnk >= 0) {
	Walln[k] = Wallnk;    	// wall number of Atom k
      }
      else {
	Wallnk     = -Wallnk;
	Walln[k]   =  Wallnk;        // loose wall number of Atom k
	LWallNu[k] = Lnu;         // mark Atom k as loose wall Atom
	looseWallCount++;
      }
      if(Wallnk >  WallN) 
	WallN =  Wallnk;
    }
    EXTRAS.close();
  }
  else {
    cout << "Ext-File \"" << ExtWallFILE << "\" not found. Not using walls and setting all masses to 1." << endl;
    for(k=0; k<N; k++) {
      Walln[k] = 0;    	// wall number of Atom k
      mi[k]    = 1.;    // inverse mass
    }
  }


  
  //printf("Number of WALLS WallN = %d. Number of loose atoms in walls = %d.\n", WallN, looseWallCount);
  cout << "Number of WALLS WallN = " << WallN << ". Number of loose atoms in walls = " << looseWallCount << endl;
  
  long WallMoveSteps;
  long WallMoveStep;
  floats Wall_Force[(WallN+1)*D];
  long Wall_AtomMax[WallN+1];
  long Wall_AtomCount[WallN+1];
  long **Wall_AtomIndex;
  Wall_AtomIndex = new long*[WallN+1];
  
  long Wall_AtomCount0;     //number of "movable" atoms
  long *Wall_AtomIndex0;
      //pointer to Wall_AtomIndex[0], to save computation time
  
  long BoundaryMoveSteps;
  long BoundaryMoveStep;
  
  long BoundaryOsciSteps;
  long BoundaryOsciStep;
  
  
  for(wi=0; wi<=WallN; wi++) {
    Wall_AtomMax[wi]   = 0;
    Wall_AtomCount[wi] = 0;
  }
  
  for(k=0; k<N; k++) 
    Wall_AtomMax[Walln[k]]++;  //number of atoms of each wall
  
  Wall_AtomIndex = new long*[WallN+1];
  for(wi=0; wi<=WallN; wi++)
    Wall_AtomIndex[wi] = new long[Wall_AtomMax[wi]];
  
  for(k=0; k<N; k++) {
    wi=Walln[k];
    Wall_AtomIndex[wi][Wall_AtomCount[wi]++]=k;
  }
  
  Wall_AtomCount0 = Wall_AtomCount[0];  //number of "movable" atoms
  Wall_AtomIndex0 = Wall_AtomIndex[0];  
	  //pointer to Wall_AtomIndex[0], to save computation time
  
    
  // ******************* read connections between atoms ******************* 

  NConns = 0;
  ifstream CONN(ConnFILE);
  if(CONN.is_open()) {
    while(!CONN.eof()){
      CONN >> j;
      CONN >> k;
      if(j<k) { conn.i1 = j-1; conn.i2 = k-1; }
      else    { conn.i1 = k-1; conn.i2 = j-1; }
      
      if( !CONN.eof() ) {
	CONN >> conn.type;
	CONN >> conn.a;
	CONN >>    dX0;
	conn.dx0 = dX0;
	dx0_to_b(dX0, conn.b, conn.type);
	
	if(CONN.peek()==13 || CONN.peek()==10) {
	  if(defaultDamping == 0.) {
	    conn.damp  = 0;
	    conn.c     = 0.;
	  }
	  else {
	    conn.damp  = 1;
	    conn.c     = defaultDamping;
	  }
	    
	  conn.c1    = 1.;
	  conn.c2    = 0.;
	  dxbb = Fcrit_to_dxbb1(conn.type, conn.a, conn.dx0, Fmax); conn.dxbb1 = dxbb; conn.dx2bb1 = (dxbb>=0. ? SQR(dxbb) : -1.);
	  dxbb = Fcrit_to_dxbb2(conn.type, conn.a, conn.dx0, Fmax); conn.dxbb2 = dxbb; conn.dx2bb2 = (dxbb>=0. ? SQR(dxbb) : -1.);
	}
	else {
	  CONN >> conn.damp;
	  CONN >> conn.c;
	  CONN >> conn.c1;
	  CONN >> conn.c2;
	  if(CONN.peek()==13 || CONN.peek()==10 || CONN.eof()) {
	    dxbb = Fcrit_to_dxbb1(conn.type, conn.a, conn.dx0, Fmax); conn.dxbb1 = dxbb; conn.dx2bb1 = (dxbb>=0. ? SQR(dxbb) : -1.);
	    dxbb = Fcrit_to_dxbb2(conn.type, conn.a, conn.dx0, Fmax); conn.dxbb2 = dxbb; conn.dx2bb2 = (dxbb>=0. ? SQR(dxbb) : -1.);
	  }
	  else {
	    CONN >> dxbb; conn.dxbb1 = dxbb; conn.dx2bb1 = (dxbb>=0. ? SQR(dxbb) : -1.);
	    CONN >> dxbb; conn.dxbb2 = dxbb; conn.dx2bb2 = (dxbb>=0. ? SQR(dxbb) : -1.);
	  }
	}
	connections.push_back(conn);
	NConns++;
      }
    }
    CONN.close();
    cout << "Number of connections read: " << NConns << endl;
  }
  else {
    cout << "Connection file \"" << ConnFILE << "\" not found. Initital number of connections: " << NConns << endl;
  }

  
  NConnsGoal = (long)floor((AvCoordNum*(floats)N)*.5 + .5);
  if(AvCoordNum >= 0. && NConns > NConnsGoal) {
    
    long z1, z2, z12ind, z12count;
    long z_i1,z_i2;
    long ParticleCoordNumMax = 0;
    long *ParticleCoordNum      = new long[N];
    long *BondIndicesCut_z12ind = new long[NConns];
    for(k=0; k<N; k++)
      ParticleCoordNum[k]=0;
    
    for(it = connections.begin(); it != connections.end(); ++it) {
      conn = *it;
      ParticleCoordNum[conn.i1]++;
      ParticleCoordNum[conn.i2]++;
    }
    
    for(k=0; k<N; k++) {
      if (ParticleCoordNum[k] > ParticleCoordNumMax)
	ParticleCoordNumMax = ParticleCoordNum[k];
    }
    
    cout << "Maximum coordination number: " << ParticleCoordNumMax << endl;
    cout << "starting to cut connections until " << NConnsGoal << " are left... " << flush;
    
    //connectionCutStruct connectionCut[NConns];
    connectionCutStruct *connectionCut = new connectionCutStruct[NConns];
    k=0;
    for(it = connections.begin(); it != connections.end(); ++it) {
      conn = *it;
      connectionCut[k].i1   = conn.i1;
      connectionCut[k].i2   = conn.i2;
      connectionCut[k].cutQ = 0;
      k++;
    }
    if(k != NConns) {
      cout << "Uhoh!" << endl;
      exit(1);
    }
      
    
    /** determine which bonds are being cut **/
    if(strcmp(cuttingMethod, "moreThan") == 0) { // "more than cuttingMoreThanCoordNum"-method
      NConns0 = NConns;
      while(NConns > NConnsGoal) {
	k = (long)floor( drand48()*(double)NConns0 );
	if( ParticleCoordNum[connectionCut[k].i1]>cuttingMoreThanCoordNum && ParticleCoordNum[connectionCut[k].i2]>cuttingMoreThanCoordNum && connectionCut[k].cutQ == 0) {
	  connectionCut[k].cutQ = 1;
	  ParticleCoordNum[connectionCut[k].i1]--;
	  ParticleCoordNum[connectionCut[k].i2]--;
	  NConns--;
	}
      }
    } 
    else if(strcmp(cuttingMethod, "highestFirst") == 0){ // "higest coordination number first"-method
      NConns0 = NConns;
      z12ind = 0;
      for(z1 = ParticleCoordNumMax; z1 > 0; z1--) {
	for(z2 = ParticleCoordNumMax; z2 >= z1; z2--) {
	  z12ind++;
	  z12count = 0;
	  for(k = 0; k < NConns0; k++) {
	    if(connectionCut[k].cutQ == 0) {
	      conn_i1 = connectionCut[k].i1; z_i1 = ParticleCoordNum[conn_i1];
	      conn_i2 = connectionCut[k].i2; z_i2 = ParticleCoordNum[conn_i2];
	      if( (z_i1 >= z1 && z_i2 >= z2) || (z_i1 >= z2 && z_i2 >= z1)  ) {
		connectionCut[k].cutQ = z12ind;
		BondIndicesCut_z12ind[z12count++] = k;
		ParticleCoordNum[conn_i1]--;
		ParticleCoordNum[conn_i2]--;
		NConns--;
	      }
	    }
	  }
	  if(NConns <= NConnsGoal) 
	    goto cuttingDone;
	}
      }
      cuttingDone:
      cout << "(loop termination at z1 = " << z1 << ", z2 = " << z2 << ")... " << flush;
      // randomly add bonds from the terminated loop, that have been cut too much
      while(NConns < NConnsGoal) {
	k = BondIndicesCut_z12ind[ (long)floor( drand48()*(double)z12count ) ];
	if(connectionCut[k].cutQ == z12ind) {
	  conn_i1 = connectionCut[k].i1; z_i1 = ParticleCoordNum[conn_i1];
	  conn_i2 = connectionCut[k].i2; z_i2 = ParticleCoordNum[conn_i2];
		    connectionCut[k].cutQ = 0;
	  ParticleCoordNum[conn_i1]++;
	  ParticleCoordNum[conn_i2]++;
	  NConns++;
	}
	else if(connectionCut[k].cutQ != 0) {
	  cout << "problem with connectionCut" << endl;
	  exit(5);
	}
      }
    }
    /** bonds to be cut were determined **/
    
    if( weakSpringsQ == 1 ) {
      k=0; i=0;
      for(it = connections.begin(); it != connections.end(); ++it) {
	if( connectionCut[k].cutQ == 1 ) {
	  (*it).a = weakSpringsa;
	  (*it).c = weakSpringsc;
	}
	else 
	  i++;


	k++;
      }
      cout << "done. Number of leftover (non-weak) connections: " << i << endl;

    }
    else {
      k=0;
      for(it = connections.begin(); it != connections.end(); ) {
	if( connectionCut[k].cutQ > 0 )
	  it = connections.erase(it); 
	else
	  ++it;

	k++;
      }
      
      k=0;
      for(it = connections.begin(); it != connections.end(); ++it)
	k++;
      
      cout << "done. Number of connections: " << k << endl;
    }

  }
  
   // ******************* read wall movements ******************* 
  WallMoveSteps = WallN>0 ? CheckNumberOfLines(WallMoveFILE) : 0;
  WallMoveStruct WallMoves[WallMoveSteps>0 ? WallMoveSteps : 1];
  
  if (WallN>0) {
    if (WallMoveSteps>0) {
      ifstream WALLMOVE(WallMoveFILE);
      for(WallMoveStep=0; WallMoveStep<WallMoveSteps; WallMoveStep++){
	

	for(nu=0;nu<D;nu++)
	  WALLMOVE >> WallMoves[WallMoveStep].direction[nu];

	WALLMOVE >> WallMoves[WallMoveStep].WallNumber;
	WALLMOVE >> WallMoves[WallMoveStep].time;
	
	if(WALLMOVE.peek()!=13 && WALLMOVE.peek()!=10 && !WALLMOVE.eof()) {
	  cout << "Wallmoves have the wrong dimension";
	  exit(1);
	}
	
	//cout << WallMoveStep << endl;
      }
      WALLMOVE.close();

      qsort (WallMoves, WallMoveSteps, sizeof(WallMoveStruct), compWM);
    }
    else {
      WallMoveSteps = 1;
      for(nu=0;nu<D;nu++)
	WallMoves[0].direction[nu] = 0.;

      WallMoves[0].WallNumber = 1;
      WallMoves[0].time       = 0.;
    }
    
    cout << "Sorted Wall Moves:" << endl;
    for(WallMoveStep=0; WallMoveStep<WallMoveSteps; WallMoveStep++) {
      cout << "at time " << WallMoves[WallMoveStep].time << ": ";
      cout << "velocity of Wall " << WallMoves[WallMoveStep].WallNumber << " changes to v = (";
      cout << WallMoves[WallMoveStep].direction[0];
      for(nu=1; nu<D; nu++)
	cout << "," << WallMoves[WallMoveStep].direction[nu];
      cout << ")" << endl;
    }
    cout << endl;
  }
  
   
     // ******************* read boundary movements ******************* 
    
  LeesEdwardsTrue   = 0;  // do not consider Lees-Edwards boundary conditions, unless set to 1 later
  BoundaryMoveSteps = CheckNumberOfLines(BoundaryMoveFILE);
  
  if (BoundaryMoveSteps==0) 
    cout << "Warning: BoundaryMoveFile " << BoundaryMoveFILE << " empty. Hence, no boundary movments." << endl;

  if (BoundaryMoveSteps<0) {
    cout << "Warning: BoundaryMoveFile " << BoundaryMoveFILE << " does not exist. Hence, no boundary movments." << endl;
    BoundaryMoveSteps = 0;
  }
  
  BoundaryMoveStruct BoundaryMoves[BoundaryMoveSteps];
  if (BoundaryMoveSteps>0) {
    
    ifstream BOUNDARYMOVE(BoundaryMoveFILE);
    
    for(BoundaryMoveStep=0; BoundaryMoveStep<BoundaryMoveSteps; BoundaryMoveStep++){
      //BOUNDARYMOVE >> bi; BoundaryMoves[BoundaryMoveStep].Number = bi;
      for(bi=0; bi<D; bi++) {
        for(nu=0; nu<D; nu++) {
          BOUNDARYMOVE >>  BoundaryMoves[BoundaryMoveStep].vel[bi][nu];
          if( (bi!=nu) && (BoundaryMoves[BoundaryMoveStep].vel[bi][nu] != 0.)) 
            LeesEdwardsTrue = 1;
        }
      }
      BOUNDARYMOVE >> BoundaryMoves[BoundaryMoveStep].time;

      if(BOUNDARYMOVE.peek()!=13 && BOUNDARYMOVE.peek()!=10 && !BOUNDARYMOVE.eof()) {
        cout << "Boundary moves have the wrong format";
        exit(1);
      }
    }
    BOUNDARYMOVE.close();
    
    /*if (LeesEdwardsTrue)
      cout << "using Lees-Edwards boundary conditions" << endl;
    else
      cout << "NOT using Lees-Edwards boundary conditions" << endl;*/

    
    qsort (BoundaryMoves, BoundaryMoveSteps, sizeof(BoundaryMoveStruct), compBM);
    
    
    cout << "Sorted Boundary Moves:" << endl;
    for(BoundaryMoveStep=0; BoundaryMoveStep<BoundaryMoveSteps; BoundaryMoveStep++) {
      cout << "time: " << BoundaryMoves[BoundaryMoveStep].time << ":" << endl;
      PrintMatrix(BoundaryMoves[BoundaryMoveStep].vel);
      /*cout << BoundaryMoves[BoundaryMoveStep].Number << " ";
      for(nu=0;nu<D;nu++)
        cout << BoundaryMoves[BoundaryMoveStep].direction[nu] << " ";*/
      cout << endl;
    }
  }
  
       // ******************* read boundary oscillations ******************* 
    
  BoundaryOsciTrue  = 0;  
  BoundaryOsciSteps = CheckNumberOfLines(BoundaryOsciFILE);
  
  if (BoundaryOsciSteps==0) 
    cout << "Warning: BoundaryOsciFile " << BoundaryOsciFILE << " empty. Hence, no boundary oscillations." << endl;

  if (BoundaryOsciSteps<0) {
    cout << "Warning: BoundaryOsciFile " << BoundaryOsciFILE << " does not exist. Hence, no boundary oscillations." << endl;
    BoundaryOsciSteps = 0;
  }
  
  BoundaryOsciStruct BoundaryOscis[BoundaryOsciSteps];
  if (BoundaryOsciSteps>0) {
    BoundaryOsciTrue = 1;
    ifstream BOUNDARYOSCI(BoundaryOsciFILE);
    
    for(BoundaryOsciStep=0; BoundaryOsciStep<BoundaryOsciSteps; BoundaryOsciStep++){
      
      for(bi=0; bi<D; bi++) {
        for(nu=0; nu<D; nu++) {
          BOUNDARYOSCI >>  BoundaryOscis[BoundaryOsciStep].amplitude[bi][nu];
          if( (bi!=nu) && (BoundaryOscis[BoundaryOsciStep].amplitude[bi][nu] != 0.)) 
            LeesEdwardsTrue = 1;
        }
      }
      for(bi=0; bi<D; bi++) 
        for(nu=0; nu<D; nu++) 
          BOUNDARYOSCI >>  BoundaryOscis[BoundaryOsciStep].frequency[bi][nu];
      
      for(bi=0; bi<D; bi++) 
        for(nu=0; nu<D; nu++) 
          BOUNDARYOSCI >>  BoundaryOscis[BoundaryOsciStep].timeShift[bi][nu];

      BOUNDARYOSCI >> BoundaryOscis[BoundaryOsciStep].time;

      if(BOUNDARYOSCI.peek()!=13 && BOUNDARYOSCI.peek()!=10 && !BOUNDARYOSCI.eof()) {
        cout << "Boundary oscillations have the wrong format";
        exit(1);
      }
    }
    BOUNDARYOSCI.close();
    
    
    qsort (BoundaryOscis, BoundaryOsciSteps, sizeof(BoundaryOsciStruct), compBO);
    
    
    cout << "Sorted Boundary Oscillations:" << endl;
    for(BoundaryOsciStep=0; BoundaryOsciStep<BoundaryOsciSteps; BoundaryOsciStep++) {
      cout << "time: " << BoundaryOscis[BoundaryOsciStep].time << ":" << endl;
      PrintOsciMatrix(BoundaryOscis[BoundaryOsciStep]);
      cout << endl;
    }
  }

  // ******************* finished reading ******************* 
  
  if (LeesEdwardsTrue)
    cout << "using Lees-Edwards boundary conditions" << endl;
  else
    cout << "NOT using Lees-Edwards boundary conditions" << endl;


  
  // ******************* start the simulation ******************* 
  cout << endl;
  cout << "Talked enough... Starting the simulation" << endl;
  cout << "========================================" << endl;
  cout << endl;
  
  count = 0;
  t = (floats)count*dt;
  
  PosOutCount    = 0;
  ForceOutCount  = 0;
  nextPosOut_t   = (floats)PosOutCount   * dt_pos;
  nextForceOut_t = (floats)ForceOutCount * dt_force;

  WallMoveStep = 0;
  BoundaryMoveStep = 0;
  BoundaryOsciStep = 0;
  
  // ******************* open streams for writing ******************* 
  ios_base::openmode mode = ios::out;
  if(continueAt >= 0) 
    mode = ios::in|ios::out;
  
  fstream BoundaryPosOutStream  (BoundaryPosOutFILE,  mode);
  fstream BoundaryForceOutStream(BoundaryForceOutFILE,mode);
  fstream WallPosOutStream      (WallPosOutFILE,      mode);
  fstream WallForceOutStream    (WallForceOutFILE,    mode);
  
  if(continueAt >= 0) {
    cout << "skipping dynamics until PosOutCount = " << continueAt << endl;
    while(PosOutCount <= continueAt) {

      if( t >= nextPosOut_t ) {
	if( PosOutCount >= continueAt ) 
	  break;
	PosOutCount++;  nextPosOut_t = (floats)PosOutCount * dt_pos;
	cout << " -- time: " << t << "\tcount: " << count << "\tnew PosOutCount: " << PosOutCount << "\tnextPosOut_t: " << nextPosOut_t << endl;
      }

      if( t >= nextForceOut_t ) {
	BoundaryPosOutStream.getline  (lineIN,LINELENGTH);
	//cout << "--- " << lineIN << endl;
	//cout << BoundaryPosOutStream.tellg() << endl;
	BoundaryForceOutStream.getline(lineIN,LINELENGTH);
	//cout << "--- " << lineIN << endl;
	//cout << BoundaryForceOutStream.tellg() << endl;
	WallPosOutStream.getline      (lineIN,LINELENGTH);
	WallForceOutStream.getline    (lineIN,LINELENGTH);

	ForceOutCount++;  nextForceOut_t = (floats)ForceOutCount * dt_force;
      }

      /******************************************\
      | update velocities of WALLS, if necessary |
      \******************************************/
      if (WallN>0) {
	while(WallMoveStep<WallMoveSteps && t>=WallMoves[WallMoveStep].time) {
	  wi = WallMoves[WallMoveStep].WallNumber;
	  
	  for(nu=0; nu<D; nu++)
	    vWall[nu] = WallMoves[WallMoveStep].direction[nu];
	  
	  for(j=0; j<Wall_AtomCount[wi]; j++) {
	    k  = Wall_AtomIndex[wi][j];
	    kD = k*D;
	    if(LWallNu[k] < 0) {
	      	  for(nu=0; nu<D; nu++)
		    V[kD+nu] = vWall[nu];
	    }
	    else {
	      for(nu=0; nu<D; nu++) {
		if( nu != LWallNu[k] )
		  V[kD+nu] = vWall[nu];
	      }
	    }
	  }
	  
	  cout << "changed vel of wall " << wi << " to v = (";
	  cout << vWall[0];
	  for(nu=1; nu<D; nu++)
	    cout << "," << vWall[nu];
	  cout << ")" <<endl;
	  WallMoveStep++;
	}
      }
      
      /***********************************************\
      | update velocities of boundaries, if necessary |
      \***********************************************/
      if (BoundaryMoveSteps>0) {
	while(BoundaryMoveStep<BoundaryMoveSteps && t>=BoundaryMoves[BoundaryMoveStep].time) {
	  //bi = BoundaryMoves[BoundaryMoveStep].Number;
	  
	  for(bi=0; bi<D; bi++)
	    for(nu=0; nu<D; nu++)
	      vBoundary[bi][nu] = BoundaryMoves[BoundaryMoveStep].vel[bi][nu];
	  

	  cout << "changing boundary velocity to: " << endl;
	  PrintMatrix(vBoundary); cout << endl;
	  BoundaryMoveStep++;
	}
      }
      
      /*************************************************\
      | update oscillations of boundaries, if necessary |
      \*************************************************/
      if (BoundaryOsciSteps>0) {
	while(BoundaryOsciStep<BoundaryOsciSteps && t>=BoundaryOscis[BoundaryOsciStep].time) {
	  
	  for(bi=0; bi<D; bi++)
	    for(nu=0; nu<D; nu++)
	      BoundaryOsci = BoundaryOscis[BoundaryMoveStep];
	  

	  cout << "changing boundary oscillation velocity to: " << endl;
	  PrintOsciMatrix(BoundaryOsci); cout << endl;
	  BoundaryOsciStep++;
	}
      }
      

      
      // update position of boundaries
      if(BoundaryMoveSteps > 0) 
	for(bi=0; bi<D; bi++) 
	  for(nu=0; nu<D; nu++) 
	    L_LE[bi][nu] += dt * vBoundary[bi][nu];
      
      if(BoundaryOsciSteps > 0) 
	for(bi=0; bi<D; bi++) 
	  for(nu=0; nu<D; nu++) 
	    L_LE[bi][nu] += dt * BoundaryOsci.amplitude[bi][nu] * cos( twoPi * BoundaryOsci.frequency[bi][nu] * (t - BoundaryOsci.timeShift[bi][nu]) );
      
      for(bi=0; bi < D; bi++) {
	L[bi]  = L_LE[bi][bi];
	Lh[bi] = L[bi]*0.5;
      }

      // update position of Lees Edwards boundaries
      if (LeesEdwardsTrue) {
	Inverse( L_LE, L_LEinv, D ); 
      }


      count++;
      t = (floats)count*dt;
    }
    
    cout << endl;
    cout << "skipping done... continuing the dynamics at: ";
    cout << "time t = " << t << "\t count = " << count << endl;
    cout << "File positions: ";
    cout << "\"" << BoundaryPosOutFILE   << "\": " << BoundaryPosOutStream.tellg()   << ";  ";
    cout << "\"" << BoundaryForceOutFILE << "\": " << BoundaryForceOutStream.tellg() << ";  ";
    cout << "\"" << WallPosOutFILE       << "\": " << WallPosOutStream.tellg()       << ";  ";
    cout << "\"" << WallForceOutFILE     << "\": " << WallForceOutStream.tellg()     << ";  ";
    cout << endl;
    cout << endl;
  }





  short firstStep = 1;
  while(t<TMAX) {
  
    /****************************************************************************************\
    | write particle positions, velocities, connections and local stress tensor if specified |
    \****************************************************************************************/
    if( t >= nextPosOut_t ) {
      if (preventOverwriteQ == 1) {
	cout << "Not overwriting output file " << addNDot(PosOutFILE,PosOutCount).c_str() << "." << endl;
	preventOverwriteQ = 0;
      }
      else {
	ofstream PosOutStream(addNDot(PosOutFILE,PosOutCount).c_str(),ios::out);
	ofstream VelOutStream(addNDot(VelOutFILE,PosOutCount).c_str(),ios::out);
	PosOutStream << setprecision(OUTPRECISION);
	VelOutStream << setprecision(OUTPRECISION);
	
	for(k=0;k<N;k++) {
	  kD=k*D;
	  PosOutStream << X[kD];
	  VelOutStream << V[kD];
	  for(nu=1;nu<D;nu++) {
	    PosOutStream << "\t" << X[kD+nu];
	    VelOutStream << "\t" << V[kD+nu];
	  }
	  PosOutStream << endl;
	  VelOutStream << endl;
	}


	PosOutStream.close(); ExecuteOutFileCmd(PosOutFILE);
	VelOutStream.close(); ExecuteOutFileCmd(VelOutFILE);
	
	
	
	// write connection file
	if( strcmp(ConnOutFILE, "none") != 0 ) {
	  ofstream ConnOutStream(addNDot(ConnOutFILE,PosOutCount).c_str(),ios::out);
	  ConnOutStream << setprecision(OUTPRECISION);
	  
	  for(it = connections.begin(); it != connections.end(); ++it) {
	    conn = *it;
	    conn.i1++;
	    conn.i2++;
	    ConnOutStream << conn.i1;
	    ConnOutStream << "\t" << conn.i2;
	    ConnOutStream << "\t" << conn.type;
	    ConnOutStream << "\t" << conn.a;
	    ConnOutStream << "\t" << conn.dx0;
	    ConnOutStream << "\t" << conn.damp;
	    ConnOutStream << "\t" << conn.c;
	    ConnOutStream << "\t" << conn.c1;
	    ConnOutStream << "\t" << conn.c2;
	    if(conn.dxbb1 >= 0. || conn.dxbb2 >= 0.) {
	      ConnOutStream << "\t" << conn.dxbb1;
	      ConnOutStream << "\t" << conn.dxbb2;
	    }
	    ConnOutStream << endl;
	  }
	  ConnOutStream.close(); ExecuteOutFileCmd(ConnOutFILE);
        }
	
	// write particle velocities relative to their neighbors
	if( strcmp(VelFluctOutFILE, "none") != 0 ) {

	  for(k=0; k<ND; k++)
	    vNeighborsSum[k] = 0.;

	  for(k=0; k<N; k++)
	    NNeighors[k] = 0;


	  ofstream VelFluctOutStream(addNDot(VelFluctOutFILE,PosOutCount).c_str(),ios::out);
	  for(it = connections.begin(); it != connections.end(); ++it) {
	    conn = *it;
	    conn_i1 = conn.i1; conn_i1D = conn_i1*D;
	    conn_i2 = conn.i2; conn_i2D = conn_i2*D;
	    NNeighors[conn_i1]++;
	    NNeighors[conn_i2]++;
	    for(nu=0; nu<D; nu++) {
	      vNeighborsSum[conn_i1D+nu] += V[conn_i2D+nu];
	      vNeighborsSum[conn_i2D+nu] += V[conn_i1D+nu];
	    }
	  }
	  
	  for(k=0; k<N; k++) {
	    kD=k*D;
	    for(nu=0; nu<D; nu++) 
	      vField[nu] = V[kD+nu] - (vNeighborsSum[kD+nu]+V[kD+nu])/(floats)(NNeighors[k]+1);


	    VelFluctOutStream << vField[0];
	    for(nu=1; nu<D; nu++) 
	      VelFluctOutStream << "\t" << vField[nu];

	    VelFluctOutStream << endl;
	  }
	  VelFluctOutStream.close(); ExecuteOutFileCmd(VelFluctOutFILE);
	}
	
	// write local stress tensor, if filename specified
	if(strcmp(SigmaOutFILE, "none") != 0) {

	  for(k=0; k<N; k++) {
	    for(nu=0; nu<D; nu++) {
	      for(bi=0; bi<D; bi++) {
		    sigmaLocalElas[k][bi][nu] = 0.;
		    sigmaLocalDiss[k][bi][nu] = 0.;
	      }
	    }
	  }

	  for(it = connections.begin(); it != connections.end(); ++it) {
	    conn = *it;
	    conn_i1  = conn.i1;
	    conn_i2  = conn.i2;
	    conn_i1D = conn.i1*D;
	    conn_i2D = conn.i2*D;
	  
	    CALC_dX0(X, conn_i1D, conn_i2D, dX);
	    CALC_dX(dX, dX_PeriodicCount);
	    dX2 = CALC_SQR_VEC(dX);
    
	    // calculate contribution of central force
	    Fh = force(conn.type, dX2, conn.a, conn.b);
	    for(nu=0; nu<D; nu++) {
	      Fnu = dX[nu]*Fh;
	      for(bi=0; bi<D; bi++) {
		sigmaLocalElas[conn_i1][bi][nu] += Fnu * dX[bi];
		sigmaLocalElas[conn_i2][bi][nu] += Fnu * dX[bi];
	      }
	    }
	  
	    // calculate contribution of damping force
	    conn_damp = conn.damp;
	    if(conn_damp > 0) {
	      CALC_dV0(V, conn_i1D, conn_i2D, dV);
	      CALC_dV(dV, dX_PeriodicCount);
	  
	      if(conn_damp == 1) {
		Fh = - conn.c;
	      }
	      else {
		dV2 = CALC_SQR_VEC(dV);
		Fh = force_damp(conn_damp, dV2, conn.c);
	      }
	  
	      if (conn.c1 != 0.) {
		c1Fh = conn.c1*Fh;
		for(nu=0; nu<D; nu++) {
		  Fnu = dV[nu]*c1Fh;
		  for(bi=0; bi<D; bi++) {
		    sigmaLocalDiss[conn_i1][bi][nu] += Fnu * dX[bi];
		    sigmaLocalDiss[conn_i2][bi][nu] += Fnu * dX[bi];
		  }
		}
	      }
	  
	      if (conn.c2 != 0.) {
		dXdV = dX[0]*dV[0];
		for(nu=1; nu<D; nu++)
		  dXdV += dX[nu]*dV[nu];
	  
		c2Fh = conn.c2 * dXdV / dX2;
		for(nu=0; nu<D; nu++) {
		  Fnu = dX[nu]*c2Fh;
		  for(bi=0; bi<D; bi++) {
		    sigmaLocalDiss[conn_i1][bi][nu] += Fnu * dX[bi];
		    sigmaLocalDiss[conn_i2][bi][nu] += Fnu * dX[bi];
		  }
		}
	      }
	    }
	  }
	  
	  // start writing
	  ofstream SigmaOutStream(addNDot(SigmaOutFILE,PosOutCount).c_str(),ios::out);
	  
	  for(k=0; k<N; k++) {

	    for (bi=0; bi<D; bi++) 
	      for (nu=0; nu<D; nu++) {
		if(bi==0 && nu==0) SigmaOutStream << 0.5*sigmaLocalElas[k][bi][nu];
		else 		 SigmaOutStream << "\t" << 0.5*sigmaLocalElas[k][bi][nu];
	      }
	
	    for (bi=0; bi<D; bi++) 
	      for (nu=0; nu<D; nu++) 
		SigmaOutStream << "\t" << 0.5*sigmaLocalDiss[k][bi][nu];

	    SigmaOutStream << endl;
	  }
	  SigmaOutStream.close(); ExecuteOutFileCmd(SigmaOutFILE);
	}
      }
      
      PosOutCount++;  nextPosOut_t = (floats)PosOutCount * dt_pos;
      cout << " -- time: " << t << "\tcount: " << count << "\tnew PosOutCount: " << PosOutCount << "\tnextPosOut_t: " << nextPosOut_t << endl;

      //cout << "L:" << endl;         PrintMatrix(L_LE);    ;
      //cout << "inverse L:" << endl; PrintMatrix(L_LEinv); cout << endl;

    }

    
    /*****************************************************\
    | calculate and write forces on Walls and boundaries  |
    \*****************************************************/
    
    /** 
    | note on indices:
    | BOUNDARIES MOVEMENT:
    | L_LE[nu][bi] is the velocity of the "bi-boundary" (boundary with normal in bi-direction) in nu-direction
    | Keep in mind: the first index is the direction, the second index the boundary.
    | e.g. L_LE[0][1] = 1. is a shear of the top-boundary one length-unit to the right.
    | By Matrix multiplication, L_LE transforms the unit square to the actual simulation box, such that
    | " (X,Y) = L_LE . (Xinv, Yinv) " yields a position (X,Y) in the simulation box, for 0 <= Xinv,Yiny < 1.
    | 
    | BOUNDARY FORCES: 
    | BoundaryForceElas[bi][nu] is the elastic contribution to the stress tensor
    | BoundaryForceDiss[bi][nu] is the dissipative contribution to the stress tensor
    | The sum of both are the actual forces
    | BoundaryForce____[bi][nu] contains the "raw" forces. 
    | To obtain the stress tensor, the forces must be normalized with 1 / ( 2 times the system size ).
    | 
    | BoundaryForce____[bi][nu] is the force on the bi-boundary in nu-direction.
    |
    | note the switched indices compared to L_LE.
    \ **/
    
    if( t >= nextForceOut_t ) {
      // forces on Walls
      //cout << " " << L[0] << "  " << L[1] << endl;
      for(wi=1;wi<=WallN;wi++) {
        wiD = wi*D;
        for(nu=0;nu<D;nu++)
          Wall_Force[wiD+nu] = 0.;
	
        for(j=0;j<Wall_AtomCount[wi];j++) {
          kD = Wall_AtomIndex[wi][j]*D;
          for(nu=0;nu<D;nu++) {
            kDpnu = kD+nu;
            Wall_Force[wiD+nu] += 0.5*(R1_F[kDpnu] + R2_F[kDpnu]);
          }
        }
	
        for(nu=0;nu<D;nu++)
          WallForceOutStream << Wall_Force[wiD+nu] << "\t";
	
      }
      WallForceOutStream << t << endl;
      
      
      // write position of first wall particle
      for(wi=1; wi<=WallN; wi++) {
        kD = Wall_AtomIndex[wi][0]*D;
        for(nu=0; nu<D; nu++)
          WallPosOutStream << X[kD+nu] << "\t";
      }
      WallPosOutStream << t << endl;
      
      
      // calculate forces on boundaries and in the whole system
      for(bi=0; bi<D; bi++) {
        for(nu=0; nu<D; nu++) {
          BoundaryForceElas[bi][nu] = 0.;
          BoundaryForceDiss[bi][nu] = 0.;
	  SystemForceElas[bi][nu]   = 0.;
          SystemForceDiss[bi][nu]   = 0.;
        }
      }
      
      
      
      
      for(it = connections.begin(); it != connections.end(); ++it) {
        conn = *it;
        conn_i1D = conn.i1*D;
        conn_i2D = conn.i2*D;
	
	CALC_Xinv(X+conn_i1D, X1_inv);
        CALC_Xinv(X+conn_i2D, X2_inv);
	
	for(bi=0; bi<D; bi++) {
	  dX_inv = X1_inv[bi] - X2_inv[bi];
	  if(dX_inv < -0.5)
	    BoundaryConnQ[bi] = -1;
	  else if(dX_inv > 0.5)
	    BoundaryConnQ[bi] = 1;
	  else
	    BoundaryConnQ[bi] = 0;
	}
	
	
	CALC_dX0(X, conn_i1D, conn_i2D, dX);
	CALC_dX(dX, dX_PeriodicCount);
	dX2 = CALC_SQR_VEC(dX);
        
	// calculate contribution of central force
	Fh = force(conn.type, dX2, conn.a, conn.b);

        for(nu=0; nu<D; nu++) {
          Fnu = dX[nu]*Fh;

	  for(bi=0; bi<D; bi++) {
	    // calculate tensor product contribution [  F (*) (R_i - R_j)  ],
	    // where "(*)" denotes the tensor product
	    SystemForceElas[bi][nu] += Fnu * dX[bi];

	    // add force contribution, if connection crosses the boundary in bi-direction
	    ADD_BOUNDARYFORCE(BoundaryForceElas, bi, nu, Fnu);
	  }
        }


        // calculate contribution of damping force
	conn_damp = conn.damp;
	if(conn_damp > 0) {
	  CALC_dV0(V, conn_i1D, conn_i2D, dV);
	  CALC_dV(dV, dX_PeriodicCount);
    
	  if(conn_damp == 1) {
	    Fh = - conn.c;
	  }
	  else {
	    dV2 = CALC_SQR_VEC(dV);
	    Fh = force_damp(conn_damp, dV2, conn.c);
	  }
    
	  if (conn.c1 != 0.) {
	    c1Fh = conn.c1*Fh;
	    for(nu=0; nu<D; nu++) {
	      Fnu = dV[nu]*c1Fh;
	      for(bi=0; bi<D; bi++) {
		SystemForceDiss[bi][nu] += Fnu * dX[bi];
		ADD_BOUNDARYFORCE(BoundaryForceDiss, bi, nu, Fnu);
	      }
	    }
	  }
    
	  if (conn.c2 != 0.) {
	    dXdV = dX[0]*dV[0];
	    for(nu=1; nu<D; nu++)
	      dXdV += dX[nu]*dV[nu];
      
	    c2Fh = conn.c2 * dXdV / dX2;
	    for(nu=0; nu<D; nu++) {
	      Fnu = dX[nu]*c2Fh;
	      for(bi=0; bi<D; bi++) {
		SystemForceDiss[bi][nu] += Fnu * dX[bi];
		ADD_BOUNDARYFORCE(BoundaryForceDiss, bi, nu, Fnu);
	      }
	    }
	  }
	}

      }
      // write forces on boundaries
      
      for (bi=0; bi<D; bi++) 
        for (nu=0; nu<D; nu++) 
          BoundaryForceOutStream << BoundaryForceElas[bi][nu] << "\t";
      
      for (bi=0; bi<D; bi++) 
        for (nu=0; nu<D; nu++) 
          BoundaryForceOutStream << BoundaryForceDiss[bi][nu] << "\t";

      for (bi=0; bi<D; bi++) 
        for (nu=0; nu<D; nu++) 
          BoundaryForceOutStream << SystemForceElas[bi][nu] << "\t";
      
      for (bi=0; bi<D; bi++) 
        for (nu=0; nu<D; nu++) 
          BoundaryForceOutStream << SystemForceDiss[bi][nu] << "\t";

	
      BoundaryForceOutStream << t << endl;

      // write boundary position
      for (bi=0; bi<D; bi++) {
        for (nu=0; nu<D; nu++) {
          if (bi == nu) 
            BoundaryPosOutStream << L[nu] << "\t";
          else
            BoundaryPosOutStream << L_LE[bi][nu] << "\t";
        }
      }
      BoundaryPosOutStream << t << endl;
      
      ForceOutCount++;  nextForceOut_t = (floats)ForceOutCount * dt_force;
    }
    
    
    /******************************************\
    | update velocities of WALLS, if necessary |
    \******************************************/
      if (WallN>0) {
	while(WallMoveStep<WallMoveSteps && t>=WallMoves[WallMoveStep].time) {
	  wi = WallMoves[WallMoveStep].WallNumber;
	  
	  for(nu=0; nu<D; nu++)
	    vWall[nu] = WallMoves[WallMoveStep].direction[nu];
	  
	  for(j=0; j<Wall_AtomCount[wi]; j++) {
	    k  = Wall_AtomIndex[wi][j];
	    kD = k*D;
	    if(LWallNu[k] < 0) {
	      	  for(nu=0; nu<D; nu++)
		    V[kD+nu] = vWall[nu];
	    }
	    else {
	      for(nu=0; nu<D; nu++) {
		if( nu != LWallNu[k] )
		  V[kD+nu] = vWall[nu];
	      }
	    }
	  }
	  
	  cout << "changed vel of wall " << wi << " to v = (";
	  cout << vWall[0];
	  for(nu=1; nu<D; nu++)
	    cout << "," << vWall[nu];
	  cout << ")" <<endl;
	  WallMoveStep++;
	}
      }
    
    /***********************************************\
    | update velocities of boundaries, if necessary |
    \***********************************************/
    if (BoundaryMoveSteps>0) {
      while(BoundaryMoveStep<BoundaryMoveSteps && t>=BoundaryMoves[BoundaryMoveStep].time) {
        //bi = BoundaryMoves[BoundaryMoveStep].Number;
	
        for(bi=0; bi<D; bi++)
          for(nu=0; nu<D; nu++)
            vBoundary[bi][nu] = BoundaryMoves[BoundaryMoveStep].vel[bi][nu];
	

        cout << "changing boundary velocity to: " << endl;
        PrintMatrix(vBoundary); cout << endl;
        BoundaryMoveStep++;
      }
    }
    
    /*************************************************\
    | update oscillations of boundaries, if necessary |
    \*************************************************/
    if (BoundaryOsciSteps>0) {
      while(BoundaryOsciStep<BoundaryOsciSteps && t>=BoundaryOscis[BoundaryOsciStep].time) {
	
        for(bi=0; bi<D; bi++)
          for(nu=0; nu<D; nu++)
            BoundaryOsci = BoundaryOscis[BoundaryMoveStep];
	

        cout << "changing boundary oscillation velocity to: " << endl;
        PrintOsciMatrix(BoundaryOsci); cout << endl;
        BoundaryOsciStep++;
      }
    }
    
    /******************************************\
    | update positions of walls and boundaries |
    \******************************************/
    
    // update positions of walls
    for(wi=1; wi<=WallN; wi++) {
      for(j=0; j<Wall_AtomCount[wi]; j++) {
	k  = Wall_AtomIndex[wi][j];
        kD = k*D;
	if(LWallNu[k] < 0) {
	  for(nu=0; nu<D; nu++) {
	    kDpnu = kD+nu;
	    X[kDpnu] += dt*V[kDpnu];
	  }
	}
	else {
	  for(nu=0; nu<D; nu++) {
	    if( nu != LWallNu[k] ) {
	      kDpnu = kD+nu;
	      X[kDpnu] += dt*V[kDpnu];
	    }
	  }
	}
      }
    }
    
    // update position of boundaries
    if(BoundaryMoveSteps > 0) 
      for(bi=0; bi<D; bi++) 
        for(nu=0; nu<D; nu++) 
          L_LE[bi][nu] += dt * vBoundary[bi][nu];
    
    if(BoundaryOsciSteps > 0) 
      for(bi=0; bi<D; bi++) 
        for(nu=0; nu<D; nu++) 
          L_LE[bi][nu] += dt * BoundaryOsci.amplitude[bi][nu] * cos( twoPi * BoundaryOsci.frequency[bi][nu] * (t - BoundaryOsci.timeShift[bi][nu]) );
    
    for(bi=0; bi < D; bi++) {
      L[bi]  = L_LE[bi][bi];
      Lh[bi] = L[bi]*0.5;
    }

     // update position of Lees Edwards boundaries
    if (LeesEdwardsTrue) {
      Inverse( L_LE, L_LEinv, D ); 
      //cout << "L:" << endl;         PrintMatrix(L_LE);    cout << endl;
      //cout << "inverse L:" << endl; PrintMatrix(L_LEinv); cout << endl;
    }


    /*****************\
    | do the dynamics |
    \*****************/

    // calculate random kicks
    if(SIGMA>0.) {
      for(j=0; j<Wall_AtomCount0; j++) {
        kD = Wall_AtomIndex0[j]*D;
        for(nu=0; nu<D; nu++) 
          Vrand[kD+nu] = randomKick();
      }
    }
    
    
    
    // choose the method
    if( strcmp(IntegrationMethod, "LeapfrogNoDamp") == 0 )  { /* Leapfrog */
      // calculate R1
      ADD_COMPONENTS(R1_X[kDpnu] = X[kDpnu] + dt_h * V[kDpnu];);
      
      calc_F(R1_X, V, connections, R1_F);
      
      // update positions of movable atoms
      ADD_COMPONENTS( V[kDpnu] = ONEminusGAMMAdt * V[kDpnu]  +  mi[k] * (dt*R1_F[kDpnu] + SIGMA*Vrand[kDpnu]);
                      X[kDpnu] = R1_X[kDpnu] + dt_h * V[kDpnu]; );

    }
    else if( strcmp(IntegrationMethod, "Leapfrog") == 0 )  { /* Leapfrog with damping*/
      if(firstStep == 1) {
        ADD_COMPONENTS( previous_X[kDpnu] = X[kDpnu]; );
      }
      
      // calculate R1
      calc_F(X, V, connections, R1_F);
      
      // calculate R2
      ADD_COMPONENTS(R1_V[kDpnu] = V[kDpnu] + mi[k]*dt*R1_F[kDpnu];
                     R2_V[kDpnu] = 0.5*( (X[kDpnu] - previous_X[kDpnu])/dt   +   R1_V[kDpnu] ); );
      
      calc_F(X, R2_V, connections, R2_F);
      
      // remember current positions for next step and 
      // update positions of movable atoms
      ADD_COMPONENTS( previous_X[kDpnu] = X[kDpnu];
                      V[kDpnu] = ONEminusGAMMAdt * V[kDpnu]  +  mi[k] * (dt*R2_F[kDpnu] + SIGMA*Vrand[kDpnu]);
                      X[kDpnu] = X[kDpnu] + dt*V[kDpnu]; );
      

    }
    else if( strcmp(IntegrationMethod, "VelocityVerlet") == 0 )  { /* Velocity Verlet with efficient damping */
      /* PSEUDO CODE: see http://wiki.vdrift.net/Old_Numerical_Integration
      
      if (not oldaccel)
        oldaccel = acceleration(state, t+dt)

      x += v*dt + 0.5*oldaccel*dt*dt
      v += 0.5*oldaccel*dt
      a = acceleration(state, t+dt)
      v += 0.5*a*dt

      oldaccel = a
      */
      if(firstStep == 1) {
	calc_F(X, V, connections, R1_F);
      }
      
      // for count>0, R1_F is remembered from the previous tmie step
      ADD_COMPONENTS( VplusAccDtH = ONEminusGAMMAdt*V[kDpnu]  +  mi[k] * (dt_h*R1_F[kDpnu] + SIGMA*Vrand[kDpnu]);
                      X[kDpnu] = X[kDpnu]  +  VplusAccDtH * dt;
                      V[kDpnu] =              VplusAccDtH;       );
      
      calc_F(X, V, connections, R2_F);
      
      
      // update velocities of movable atoms and
      // remember current forces (R1_F) for next step 
      ADD_COMPONENTS( V[kDpnu] = ONEminusGAMMAdt*V[kDpnu]  +  mi[k] * (dt_h*R2_F[kDpnu] + SIGMA*Vrand[kDpnu]); 
                      R1_F[kDpnu] = R2_F[kDpnu]; );


    }
    else if( strcmp(IntegrationMethod, "Euler") == 0 )  {  /* Euler integration */
      // calculate R1 
      calc_F(X, V, connections, R1_F);
      
      // update positions of movable atoms
      ADD_COMPONENTS( X[kDpnu] = X[kDpnu] + dt*V[kDpnu];
                      V[kDpnu] = ONEminusGAMMAdt * V[kDpnu]  +  mi[k] * (dt*R1_F[kDpnu] + SIGMA*Vrand[kDpnu]); );
      
    }
    else if( strcmp(IntegrationMethod, "Overdamped") == 0 ){  /* Overdamped Heun method */
      
      // calculate R1 using X and V
      calc_F(X, V, connections, R1_F);
      ADD_COMPONENTS( R1_V[kDpnu] = GAMMAi * R1_F[kDpnu];
		      R1_X[kDpnu] = X[kDpnu] + dt*R1_V[kDpnu] + SIGMAoGAMMA*Vrand[kDpnu]; );
      
      // calculate R2 using R1
      calc_F(R1_X, R1_V, connections, R2_F);
      ADD_COMPONENTS( R2_V[kDpnu] = GAMMAi * R2_F[kDpnu];
		      /*R2_X[kDpnu] = X[kDpnu] + dth*( R1_V[kDpnu] + R2_V[kDpnu]) + SIGMA*Vrand[kDpnu]; */);

      
      // update positions of movable atoms
      ADD_COMPONENTS( V[kDpnu] = 0.5*(R1_V[kDpnu] + R2_V[kDpnu]);
                      X[kDpnu] = X[kDpnu] + dt*V[kDpnu] + SIGMAoGAMMA*Vrand[kDpnu]; );

    }
    else if( strcmp(IntegrationMethod, "Heun") == 0 ) {  /* Default algorithm: Heun method */
      
      // calculate R1 using X and V
      calc_F(X, V, connections, R1_F);
      ADD_COMPONENTS( R1_X[kDpnu] = X[kDpnu] + dt_h*V[kDpnu];
                      R1_V[kDpnu] = ONEminusGAMMAdt_h * V[kDpnu]  +  mi[k] * (dt_h*R1_F[kDpnu] + SIGMA*Vrand[kDpnu]); );
      
      // calculate R2 using R1
      calc_F(R1_X, R1_V, connections, R2_F);
      ADD_COMPONENTS( R2_X[kDpnu] = R1_X[kDpnu] + dt_h*R1_V[kDpnu];
                      R2_V[kDpnu] = ONEminusGAMMAdt_h * R1_V[kDpnu]  +  mi[k] * (dt_h*R2_F[kDpnu] + SIGMA*Vrand[kDpnu]); );
      
      // update positions of movable atoms
      ADD_COMPONENTS( X[kDpnu] = 0.5*(R1_X[kDpnu] + R2_X[kDpnu]);
                      V[kDpnu] = 0.5*(R1_V[kDpnu] + R2_V[kDpnu]); );

    }
    else {
      cout << "Integration method " << IntegrationMethod << " unknown!" << endl;
      exit(10);
    }


    /*****************************\
    | remove and add interactions |
    \*****************************/
    
    // remove interactions
    if (RemoveProb > 0.) {
      for(it = connections.begin(); it != connections.end(); ) {
        //if( (*it).type==1 ) {
          conn = *it;
          conn_i1D = conn.i1*D;
          conn_i2D = conn.i2*D;
          
          CALC_dX0(X, conn_i1D, conn_i2D, dX);
          CALC_dX(dX, dX_PeriodicCount);
          dX2 = CALC_SQR_VEC(dX);
          
          if( dX2 > dX2Remove || dX2 < conn.dx2bb1 || (conn.dx2bb2 > 0. && dX2 > conn.dx2bb2) || drand48()<RemoveProb )
            it = connections.erase(it); 
          else
            ++it;
      }
    }
    else {
      // RemoveProb=0  --> no need to spend time on random number generation:
      for(it = connections.begin(); it != connections.end(); ) 	{
        conn = *it;
        conn_i1D = conn.i1*D;
        conn_i2D = conn.i2*D;
          
        CALC_dX0(X, conn_i1D, conn_i2D, dX);
        CALC_dX(dX, dX_PeriodicCount);
        dX2 = CALC_SQR_VEC(dX);
        // Fh = force((*it).type, dX2, (*it).a, (*it).b);
        
        if( dX2 > dX2Remove || dX2 < conn.dx2bb1 || (conn.dx2bb2 > 0. && dX2 > conn.dx2bb2) )
          it = connections.erase(it); 
        else
          ++it;
  
      }
    }
    
    // add interactions
    for(i=0; i<AddNumber; i++) {
      j = (long)floor( drand48()*(double)N ); conn_i1D=j*D;
      k = (long)floor( drand48()*(double)N ); conn_i2D=k*D;
      if (j!=k) {
        
        CALC_dX0(X, conn_i1D, conn_i2D, dX);
        CALC_dX(dX, dX_PeriodicCount);
        dX2 = CALC_SQR_VEC(dX);
	
	if(dX2 < dX2Add) {
	  if(j<k) {
	    conn.i1 = j;
	    conn.i2 = k;
	  }
	  else {
    	    conn.i1 = k;
	    conn.i2 = j;
	  }
	  conn.type = 5;
          conn.damp = 0;
	  conn.a    = 1.;
          conn.dx0  = 1.;
          dx0_to_b(conn.dx0, conn.b, conn.type);
          conn.c    = 0.;
          conn.c1   = 0.;
          conn.c2   = 0.;
          conn.dxbb1 = -1.;
          conn.dxbb2 = -1.;
	  connections.push_back(conn);
	}
      }
    }
	

    
    //cout << "time: " << t << " pos: " << X[0] << " " << X[1] << endl;
    count++;
    t = (floats)count*dt;
    
    if(firstStep == 1)
      firstStep = 0;
  }
  


  
  // free memory
  for (wi=0; wi <= WallN; wi++)
    delete[] Wall_AtomIndex[wi];
  delete[] Wall_AtomIndex;
  
  // close writing streams
  BoundaryPosOutStream.close();
  BoundaryForceOutStream.close();
  WallPosOutStream.close();
  WallForceOutStream.close();
}
