"""
Input:
       position and velocity files of all time
Output:
       fig: front position - time, for various Vy threshold
       txt: same above with separate files for each Vy
calculate the front position in 1D lattice spring model
according to the transverse velocity: Vy
front position is the position of the bead whose absolute Vy is below the critical value vc
"""

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.cm as cm 
import sys
import os
import time

#--------------
numberinitial = 1  # file batch number
numberfinal = 50
numberinterval = 1
dt_pos=100 # should be the same as in Settingxxx.txt in MD config file
data_range_min=0 # to use these range of data
data_range_max=9999
flagmax=10 # number of continuous points satisfying the velocity condition of front

vcmin=1e-8
vcmax=1.5e-8
vcinterval=1e-8

indir='5e0/'
indir2='damp0/run/'
outdir='front/' #output dir
if not os.path.exists(outdir):
    os.makedirs(outdir)

f2=indir+'BOUNDARYmlattice.txt'
vel=np.loadtxt(f2)
v0=vel[2]*0.5
print 'v0 =',v0
txtname='to be defined later'
figname='detail_front_shear rate: %e.png'% v0

#test final file existence
f1=indir+indir2+'OUTx%d.txt' % numberfinal
pos = np.loadtxt(f1)
#--------------

sizeofarray=(numberfinal+numberinterval-numberinitial)/numberinterval


vc=vcmin
while vc<=vcmax:
    i=0
    t=np.zeros(sizeofarray)
    xfront=np.zeros(sizeofarray)
    
    for number in range(numberinitial, numberfinal+numberinterval, numberinterval):
        percentage=100.0*(number-numberinitial)/(numberfinal-numberinitial)
        sys.stdout.write('\r%d%%'% percentage)
        sys.stdout.flush()

        f1=indir+indir2+'OUTx%d.txt' % number # set input file name
        f2=indir+indir2+'OUTv%d.txt' % number


        pos = np.loadtxt(f1) # read arrays of position and velocity from the file
        vel = np.loadtxt(f2) 
        x = pos[:,0] # separate x and y to different array
        y = pos[:,1]
        velx=vel[:,0]
        vely=vel[:,1]

        t[i]=(numberinitial+i*numberinterval)*dt_pos # cal. the time

        flag=0 # to be compared with flagmax
        for data_num in range(data_range_min,data_range_max+1,1):
            if (abs(vely[data_num])<vc):
                #   print (abs(vely[data_num]),vc) #debug
                #  time.sleep(3) #debug
                flag+=1
            if flag==flagmax:
            #    print x[data_num-flagmax+1], data_num-flagmax+1 #debug
            #    time.sleep(3) #debug
                xfront[i]=x[data_num-flagmax+1]
                break

        if flag==0:
            print('flag==0, i=%d'% i)
            break

        i+=1         

    k=int(round((vc-vcmin)/vcinterval))
#    print k
    k255=int(round(255*(vc-vcmin)/(vcmax-vcmin)))
    txtname='front_shear rate:%e_shreshold %1.1e.txt'% (v0,vc)
    np.savetxt(outdir+txtname,np.column_stack((t,xfront)))
    plt.scatter(t,xfront,s=50,c=cm.gist_rainbow(k255),label='%1.1e'% vc)
  
    plt.legend(loc='upper left',title='Shear rate: '+str(v0)+'\nVy threshold for front')
    vc+=vcinterval
    
plt.xlabel('Time')
plt.ylabel('Front Position')
plt.savefig(outdir+figname)
plt.show()


